/*
 * XBoard -- an Xt/Athena user interface for GNU Chess
 *
 * Original authors:  Dan Sears and Chris Sears
 * Enhancements (Version 2.0):  Tim Mann
 *
 * XBoard borrows its colors, icon and piece bitmaps from XChess
 * which was written and is copyrighted by Wayne Christopher.
 *
 * Copyright 1991 by Digital Equipment Corporation, Maynard, Massachusetts.
 * Enhancements Copyright 1992 Free Software Foundation, Inc.
 *
 * The following terms apply to Digital Equipment Corporation's copyright
 * interest in XBoard:
 * ------------------------------------------------------------------------
 * All Rights Reserved
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of Digital not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * ------------------------------------------------------------------------
 *
 * The following terms apply to the enhanced version of XBoard distributed
 * by the Free Software Foundation:
 * ------------------------------------------------------------------------
 * This file is part of XBOARD.
 *
 * XBOARD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY.  No author or distributor accepts responsibility to anyone for
 * the consequences of using it or for whether it serves any particular
 * purpose or works at all, unless he says so in writing.  Refer to the XBOARD
 * General Public License for full details.
 *
 * Everyone is granted permission to copy, modify and redistribute XBOARD, but
 * only under the conditions described in the XBOARD General Public License. A
 * copy of this license is supposed to have been given to you along with
 * XBOARD so you can know your rights and responsibilities.  It should be in a
 * file named COPYING.  Among other things, the copyright notice and this
 * notice must be preserved on all copies.
 * ------------------------------------------------------------------------
 *
 * Revision 1.0 90/10/31
 *		Initial release.
 *
 * Revision 1.1 91/01/26
 *		Major bug fix release.
 *
 * Revision 1.2 91/06/11
 *		Another major bug fix release.
 *		lex game file parser.
 *		Popup dialogs for file names.
 *		SYSV support.
 *
 * Revision 2.0 92/05/19, Tim Mann
 *              Countless bug fixes and small improvements.
 *              Full support for underpromotion.
 *              Mode indicator.
 *              Edit Position mode.
 */

#define VERSION "2.0"

#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#ifdef	__STDC__
#include <stdlib.h>
#endif
#if	SYSTEM_FIVE || SYSV
#include <sys/types.h>
#include <sys/stat.h>
#endif
#include <pwd.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/cursorfont.h>
#include "xboard.h"
#include "patchlevel.h"

#include "bitmaps/s_p.bm"
#include "bitmaps/s_r.bm"
#include "bitmaps/s_kt.bm"
#include "bitmaps/s_b.bm"
#include "bitmaps/s_q.bm"
#include "bitmaps/s_k.bm"

#include "bitmaps/ol_p.bm"
#include "bitmaps/ol_r.bm"
#include "bitmaps/ol_kt.bm"
#include "bitmaps/ol_b.bm"
#include "bitmaps/ol_q.bm"
#include "bitmaps/ol_k.bm"

#include "bitmaps/sm_s_p.bm"
#include "bitmaps/sm_s_r.bm"
#include "bitmaps/sm_s_kt.bm"
#include "bitmaps/sm_s_b.bm"
#include "bitmaps/sm_s_q.bm"
#include "bitmaps/sm_s_k.bm"

#include "bitmaps/sm_ol_p.bm"
#include "bitmaps/sm_ol_r.bm"
#include "bitmaps/sm_ol_kt.bm"
#include "bitmaps/sm_ol_b.bm"
#include "bitmaps/sm_ol_q.bm"
#include "bitmaps/sm_ol_k.bm"

#include "bitmaps/xs_s_p.bm"
#include "bitmaps/xs_s_r.bm"
#include "bitmaps/xs_s_kt.bm"
#include "bitmaps/xs_s_b.bm"
#include "bitmaps/xs_s_q.bm"
#include "bitmaps/xs_s_k.bm"

#include "bitmaps/xs_ol_p.bm"
#include "bitmaps/xs_ol_r.bm"
#include "bitmaps/xs_ol_kt.bm"
#include "bitmaps/xs_ol_b.bm"
#include "bitmaps/xs_ol_q.bm"
#include "bitmaps/xs_ol_k.bm"

#include "bitmaps/icon.bm"

void main P((int argc, char **argv));
void CreateGCs P((void));
void CreatePieces P((void));
void CreatePieceMenus P((void));
void PieceMenuPopup P((Widget w, XEvent *event, String *params,
			Cardinal *num_params));
static void PieceMenuSelect P((Widget w, ChessSquare piece, caddr_t junk));
static void SetWhiteToPlay P((void));
static void SetBlackToPlay P((void));
void ReadBitmap P((String name, Pixmap *pm,
	char large_bits[], char medium_bits[], char small_bits[]));
void CreateGrid P((void));
int EventToSquare P((int x));
ChessSquare CharToPiece P((int c));
void DrawSquare P((int row, int column, ChessSquare piece));
void EventProc P((Widget widget, caddr_t unused, XEvent *event));
void DrawPosition P((Widget w, XExposeEvent *event));
void InitPosition P((void));
void CopyBoard P((Board to, Board from));
void SendCurrentBoard P((FILE *fp));
void SendBoard P((FILE *fp, Board board));
void HandleUserMove P((Widget w, XEvent *event));
void FinishUserMove P((ChessMove move_type, int to_x, int to_y));
void HandleMachineMove P((char *message, FILE *fp));
void ReadGameFile P((void));
int ReadGameFileProc P((void));
void MakeMove P((ChessMove *move_type, int from_x, int from_y,
	int to_x, int to_y));
void InitChessProgram P((char *host_name, char *program_name, int *pid,
	FILE **to, FILE **from, XtIntervalId *xid));
void ShutdownChessPrograms P((char *message));
void CommentPopUp P((char *label));
void FileNamePopUp P((char *label, void (*proc)(char *name)));
void FileNameCallback P((Widget w, XtPointer client_data,
			 XtPointer call_data));
void FileNameAction P((Widget w, XEvent *event));
void PromotionPopUp P((ChessSquare piece, int to_x, int to_y));
void PromotionCallback P((Widget w, XtPointer client_data,
			  XtPointer call_data));
void SelectCommand P((Widget w, XtPointer client_data, XtPointer call_data));
void ModeHighlight P((void));
void GameProc P((void));
void QuitProc P((void));
int PlayFromGameFileProc P((char *name));
void MachinePlaysBlackProc P((void));
void ForwardProc P((void));
void ResetFileProc P((void));
void ResetProc P((void));
int SetupPositionFromFileProc P((char *name));
void MachinePlaysWhiteProc P((void));
void BackwardProc P((void));
void FlipProc P((void));
void SaveGameProc P((char *name));
void SwitchProc P((void));
void EditPositionProc P((void));
void EditPositionDone P((void));
void ForceProc P((void));
void HintProc P((void));
void SavePositionProc P((char *name));
void TwoMachinesPlayProc P((void));
void PauseProc P((void));
void Iconify P((void));
void SendToProgram P((char *message, FILE *fp));
void ReceiveFromProgram P((FILE *fp, int *source, XtInputId *id));
void DisplayMessage P((char *message));
void DisplayName P((char *name));
void Attention P((int pid));
void DisplayClocks P((int clock_mode));
void DisplayTimerLabel P((Widget w, char *color, time_t timer));
char *TimeString P((time_t tm));
void Usage P((void));
char *StrStr P((char *string, char *match));
#if	SYSTEM_FIVE || SYSV
char *PseudoTTY P((int *ptyv));
#else
void CatchPipeSignal P((void));
#endif
extern int yylex P((void));
extern int yynewfile P((void));

/*
 * XBoard depends on Xt R4 or higher
 */
int xtVersion = XtSpecificationRelease;

int xScreen;
Display *xDisplay;
Window xBoardWindow;
GC lightSquareGC, darkSquareGC, lineGC, wdPieceGC, wlPieceGC,
	bdPieceGC, blPieceGC, wbPieceGC, bwPieceGC;
Pixmap solidPawnBitmap, solidRookBitmap, solidKnightBitmap,
        solidBishopBitmap, solidQueenBitmap, solidKingBitmap,
        outlinePawnBitmap, outlineRookBitmap, outlineKnightBitmap,
        outlineBishopBitmap, outlineQueenBitmap, outlineKingBitmap, iconPixmap;
Widget shellWidget, formWidget, boardWidget, commandsWidget, messageWidget,
	whiteTimerWidget, blackTimerWidget, nameWidget, widgetList[6], 
        commentShell, promotionShell, whitePieceMenu, blackPieceMenu;
XSegment gridSegments[(BOARD_SIZE + 1) * 2];
XtIntervalId firstProgramXID = NULL, secondProgramXID = NULL,
	readGameXID = NULL, timerXID = NULL;
XFontStruct *labelFont;
XtAppContext appContext;
XtCallbackProc fileProc;

FILE *fromFirstProgFP, *toFirstProgFP, *fromSecondProgFP,
	*toSecondProgFP, *gameFileFP, *lastMsgFP;
int currentMove = 0, forwardMostMove = 0, firstProgramPID = 0,
	secondProgramPID = 0, squareSize = LARGE_SQUARE_SIZE, fromX = -1,
	fromY = -1, firstMove = True, flipView = False,
	xboardDebug, commentUp = False, filenameUp = False,
        blackPlaysFirst = False, startedFromSetupPosition = False,
	promotionUp = False, timeControl = 0, pmFromX = -1, pmFromY = -1;
unsigned long timerForegroundPixel, timerBackgroundPixel;
MatchMode matchMode = MatchFalse;
GameMode gameMode = BeginningOfGame, lastGameMode = BeginningOfGame;
BoardSize boardSize = Large;
char moveList[MAX_MOVES][MOVE_LEN], parseList[MAX_MOVES][MOVE_LEN * 2],
	ptyname[24], *ttyname, *chessDir, *programName;

time_t whiteTimeRemaining, blackTimeRemaining;
extern char currentMoveString[];
extern char yytext[];

Board boards[MAX_MOVES], initialPosition = {
	{ WhiteRook, WhiteKnight, WhiteBishop, WhiteQueen,
		WhiteKing, WhiteBishop, WhiteKnight, WhiteRook },
	{ WhitePawn, WhitePawn, WhitePawn, WhitePawn,
		WhitePawn, WhitePawn, WhitePawn, WhitePawn },
	{ EmptySquare, EmptySquare, EmptySquare, EmptySquare,
		EmptySquare, EmptySquare, EmptySquare, EmptySquare },
	{ EmptySquare, EmptySquare, EmptySquare, EmptySquare,
		EmptySquare, EmptySquare, EmptySquare, EmptySquare },
	{ EmptySquare, EmptySquare, EmptySquare, EmptySquare,
		EmptySquare, EmptySquare, EmptySquare, EmptySquare },
	{ EmptySquare, EmptySquare, EmptySquare, EmptySquare,
		EmptySquare, EmptySquare, EmptySquare, EmptySquare },
	{ BlackPawn, BlackPawn, BlackPawn, BlackPawn,
		BlackPawn, BlackPawn, BlackPawn, BlackPawn },
	{ BlackRook, BlackKnight, BlackBishop, BlackQueen,
		BlackKing, BlackBishop, BlackKnight, BlackRook }
};

String buttonStrings[] = {
	"Quit", "Load Game", "Machine Black", "Forward",
	"Reset", "Load Position", "Machine White", "Backward",
	"Flip View", "Save Game", "Force Moves", "Edit Position",
	"Hint", "Save Position", "Two Machines", "Pause"
};
/* must be in same order as buttonStrings! */
typedef enum {
	ButtonQuit, ButtonLoadGame, ButtonMachineBlack, ButtonForward,
	ButtonReset, ButtonLoadPosition, ButtonMachineWhite, ButtonBackward,
	ButtonFlipView, ButtonSaveGame, ButtonForceMoves, ButtonEditPosition,
	ButtonHint, ButtonSavePosition, ButtonTwoMachines, ButtonPause
} Button;

#define PIECE_MENU_SIZE 10
String pieceMenuStrings[PIECE_MENU_SIZE] = {
	"----", "Pawn", "Knight", "Bishop", "Rook", "Queen", "King",
	"----", "Empty square", "Clear board"
};
/* must be in same order as PieceMenuStrings! */
ChessSquare pieceMenuTranslation[2][PIECE_MENU_SIZE] = {
	{ (ChessSquare) 0, WhitePawn, WhiteKnight, WhiteBishop,
	    WhiteRook, WhiteQueen, WhiteKing,
            (ChessSquare) 0, EmptySquare, ClearBoard },
	{ (ChessSquare) 0, BlackPawn, BlackKnight, BlackBishop,
	    BlackRook, BlackQueen, BlackKing,
            (ChessSquare) 0, EmptySquare, ClearBoard },
};

Arg shellArgs[] = {
	{ XtNwidth, 0 },
	{ XtNheight, 0 },
	{ XtNminWidth, 0 },
	{ XtNminHeight, 0 },
	{ XtNmaxWidth, 0 },
	{ XtNmaxHeight, 0 }
};

Arg boardArgs[] = {
	{ XtNborderWidth, 0 },
	{ XtNwidth, LINE_GAP + BOARD_SIZE * (LARGE_SQUARE_SIZE + LINE_GAP) },
	{ XtNheight, LINE_GAP + BOARD_SIZE * (LARGE_SQUARE_SIZE + LINE_GAP) }
};

Arg commandsArgs[] = {
	{ XtNborderWidth, 0 },
	{ XtNdefaultColumns, 4 },
	{ XtNforceColumns, True },
	{ XtNlist, (int) buttonStrings },
	{ XtNnumberStrings, XtNumber(buttonStrings) }
};

Arg messageArgs[] = {
	{ XtNborderWidth, 0 },
	{ XtNjustify, (int) XtJustifyLeft },
	{ XtNlabel, (int) "starting..." }
};

Arg timerArgs[] = {
	{ XtNborderWidth, 0 },
	{ XtNjustify, (int) XtJustifyLeft }
};

Arg nameArgs[] = {
	{ XtNborderWidth, 0 },
	{ XtNjustify, (int) XtJustifyLeft }
};

typedef struct {
	Pixel whitePieceColor;
	Pixel blackPieceColor;
	Pixel lightSquareColor;
	Pixel darkSquareColor;
	int movesPerSession;
	String initString;
	String whiteString;
	String blackString;
	String firstChessProgram;
	String secondChessProgram;
	Boolean noChessProgram;
	String firstHost;
	String secondHost;
	String solidPawnBitmap;
	String solidRookBitmap;
	String solidBishopBitmap;
	String solidKnightBitmap;
	String solidQueenBitmap;
	String solidKingBitmap;
	String outlinePawnBitmap;
	String outlineRookBitmap;
	String outlineBishopBitmap;
	String outlineKnightBitmap;
	String outlineQueenBitmap;
	String outlineKingBitmap;
	String remoteShell;
	float timeDelay;
	String timeControl;
	String loadGameFile;
	String loadPositionFile;
	String saveGameFile;
	String savePositionFile;
	String matchMode;
	Boolean monoMode;
	Boolean debugMode;
	Boolean clockMode;
	String boardSize;
	Boolean Iconic;
	String searchTime;
	int searchLevel;
} AppData, *AppDataPtr;

AppData appData;

XtResource clientResources[] = {
	{
		"whitePieceColor", "WhitePieceColor", XtRPixel, sizeof(Pixel),
		XtOffset(AppDataPtr, whitePieceColor), XtRString,
		WHITE_PIECE_COLOR
	}, {
		"blackPieceColor", "BlackPieceColor", XtRPixel, sizeof(Pixel),
		XtOffset(AppDataPtr, blackPieceColor), XtRString,
		BLACK_PIECE_COLOR
	}, {
		"lightSquareColor", "LightSquareColor", XtRPixel,
		sizeof(Pixel), XtOffset(AppDataPtr, lightSquareColor),
		XtRString, LIGHT_SQUARE_COLOR
	}, {
		"darkSquareColor", "DarkSquareColor", XtRPixel, sizeof(Pixel),
		XtOffset(AppDataPtr, darkSquareColor), XtRString,
		DARK_SQUARE_COLOR
	}, {
		"movesPerSession", "movesPerSession", XtRInt, sizeof(int),
		XtOffset(AppDataPtr, movesPerSession), XtRImmediate,
		(XtPointer) MOVES_PER_SESSION
	}, {
		"initString", "initString", XtRString, sizeof(String),
		XtOffset(AppDataPtr, initString), XtRString, INIT_STRING
	}, {
		"whiteString", "whiteString", XtRString, sizeof(String),
		XtOffset(AppDataPtr, whiteString), XtRString, WHITE_STRING
	}, {
		"blackString", "blackString", XtRString, sizeof(String),
		XtOffset(AppDataPtr, blackString), XtRString, BLACK_STRING
	}, {
		"firstChessProgram", "firstChessProgram", XtRString,
		sizeof(String), XtOffset(AppDataPtr, firstChessProgram),
		XtRString, FIRST_CHESS_PROGRAM
	}, {
		"secondChessProgram", "secondChessProgram", XtRString,
		sizeof(String), XtOffset(AppDataPtr, secondChessProgram),
		XtRString, SECOND_CHESS_PROGRAM
	}, {
		"noChessProgram", "noChessProgram", XtRBoolean,
		sizeof(Boolean), XtOffset(AppDataPtr, noChessProgram),
		XtRImmediate, (XtPointer) False
	}, {
		"firstHost", "firstHost", XtRString, sizeof(String),
		XtOffset(AppDataPtr, firstHost), XtRString, FIRST_HOST
	}, {
		"secondHost", "secondHost", XtRString, sizeof(String),
		XtOffset(AppDataPtr, secondHost), XtRString, SECOND_HOST
	}, {
		"solidPawnBitmap", "solidPawnBitmap", XtRString,
		sizeof(String), XtOffset(AppDataPtr, solidPawnBitmap),
		XtRString, NULL
	}, {
		"solidRookBitmap", "solidRookBitmap", XtRString,
		sizeof(String), XtOffset(AppDataPtr, solidRookBitmap),
		XtRString, NULL
	}, {
		"solidKnightBitmap", "solidKnightBitmap", XtRString,
		sizeof(String), XtOffset(AppDataPtr, solidKnightBitmap),
		XtRString, NULL
	}, {
		"solidBishopBitmap", "solidBishopBitmap", XtRString,
		sizeof(String), XtOffset(AppDataPtr, solidBishopBitmap),
		XtRString, NULL
	}, {
		"solidQueenBitmap", "solidQueenBitmap", XtRString,
		sizeof(String), XtOffset(AppDataPtr, solidQueenBitmap),
		XtRString, NULL
	}, {
		"solidKingBitmap", "solidKingBitmap", XtRString,
		sizeof(String), XtOffset(AppDataPtr, solidKingBitmap),
		XtRString, NULL
	}, {
		"outlinePawnBitmap", "outlinePawnBitmap", XtRString,
		sizeof(String), XtOffset(AppDataPtr, outlinePawnBitmap),
		XtRString, NULL
	}, {
		"outlineRookBitmap", "outlineRookBitmap", XtRString,
		sizeof(String), XtOffset(AppDataPtr, outlineRookBitmap),
		XtRString, NULL
	}, {
		"outlineKnightBitmap", "outlineKnightBitmap", XtRString,
		sizeof(String), XtOffset(AppDataPtr, outlineKnightBitmap),
		XtRString, NULL
	}, {
		"outlineBishopBitmap", "outlineBishopBitmap", XtRString,
		sizeof(String), XtOffset(AppDataPtr, outlineBishopBitmap),
		XtRString, NULL
	}, {
		"outlineQueenBitmap", "outlineQueenBitmap", XtRString,
		sizeof(String), XtOffset(AppDataPtr, outlineQueenBitmap),
		XtRString, NULL
	}, {
		"outlineKingBitmap", "outlineKingBitmap", XtRString,
		sizeof(String), XtOffset(AppDataPtr, outlineKingBitmap),
		XtRString, NULL
	}, {
		"remoteShell", "remoteShell", XtRString, sizeof(String),
		XtOffset(AppDataPtr, remoteShell), XtRString, "rsh"
	}, {
		"timeDelay", "timeDelay", XtRFloat, sizeof(float),
		XtOffset(AppDataPtr, timeDelay), XtRString,
		(XtPointer) TIME_DELAY
	}, {
		"timeControl", "timeControl", XtRString, sizeof(String),
		XtOffset(AppDataPtr, timeControl), XtRString,
		(XtPointer) TIME_CONTROL
	}, {
		"loadGameFile", "loadGameFile", XtRString, sizeof(String),
		XtOffset(AppDataPtr, loadGameFile), XtRString, NULL
	}, {
		"loadPositionFile", "loadPositionFile", XtRString,
		sizeof(String), XtOffset(AppDataPtr, loadPositionFile),
		XtRString, NULL
	}, {
		"saveGameFile", "saveGameFile", XtRString, sizeof(String),
		XtOffset(AppDataPtr, saveGameFile), XtRString, ""
	}, {
		"savePositionFile", "savePositionFile", XtRString,
		sizeof(String), XtOffset(AppDataPtr, savePositionFile),
		XtRString, ""
	}, {
		"matchMode", "matchMode", XtRString, sizeof(String),
		XtOffset(AppDataPtr, matchMode), XtRString, MATCH_MODE
	}, {
		"monoMode", "monoMode", XtRBoolean, sizeof(Boolean),
		XtOffset(AppDataPtr, monoMode), XtRImmediate,
		(XtPointer) False
	}, {
		"debugMode", "debugMode", XtRBoolean, sizeof(Boolean),
		XtOffset(AppDataPtr, debugMode), XtRImmediate,
		(XtPointer) False
	}, {
		"Iconic", "Iconic", XtRBoolean, sizeof(Boolean),
		XtOffset(AppDataPtr, Iconic), XtRImmediate,
		(XtPointer) False
	}, {
		"clockMode", "clockMode", XtRBoolean, sizeof(Boolean),
		XtOffset(AppDataPtr, clockMode), XtRImmediate,
		(XtPointer) True
	}, {
		"boardSize", "boardSize", XtRString, sizeof(String),
		XtOffset(AppDataPtr, boardSize), XtRString, DEFAULT_SIZE
	}, {
		"searchTime", "searchTime", XtRString, sizeof(String),
		XtOffset(AppDataPtr, searchTime), XtRString,
		(XtPointer) NULL
	}, {
		"searchLevel", "searchLevel", XtRInt, sizeof(int),
		XtOffset(AppDataPtr, searchLevel), XtRImmediate, 
		(XtPointer) 0
	}
};

Pixmap *pieceToSolid[] = {
	&solidPawnBitmap, &solidRookBitmap, &solidKnightBitmap,
	&solidBishopBitmap, &solidQueenBitmap, &solidKingBitmap,
	&solidPawnBitmap, &solidRookBitmap, &solidKnightBitmap,
	&solidBishopBitmap, &solidQueenBitmap, &solidKingBitmap
};

Pixmap *pieceToOutline[] = {
	&outlinePawnBitmap, &outlineRookBitmap, &outlineKnightBitmap,
	&outlineBishopBitmap, &outlineQueenBitmap, &outlineKingBitmap,
	&outlinePawnBitmap, &outlineRookBitmap, &outlineKnightBitmap,
	&outlineBishopBitmap, &outlineQueenBitmap, &outlineKingBitmap
};

char pieceToChar[] = {
	'P', 'R', 'N', 'B', 'Q', 'K',
	'p', 'r', 'n', 'b', 'q', 'k', '.'
};

XrmOptionDescRec shellOptions[] = {
	{ "-whitePieceColor", "whitePieceColor", XrmoptionSepArg, NULL },
	{ "-wpc", "whitePieceColor", XrmoptionSepArg, NULL },
	{ "-blackPieceColor", "blackPieceColor", XrmoptionSepArg, NULL },
	{ "-bpc", "blackPieceColor", XrmoptionSepArg, NULL },
	{ "-lightSquareColor", "lightSquareColor", XrmoptionSepArg, NULL },
	{ "-lsc", "lightSquareColor", XrmoptionSepArg, NULL },
	{ "-darkSquareColor", "darkSquareColor", XrmoptionSepArg, NULL },
	{ "-dsc", "darkSquareColor", XrmoptionSepArg, NULL },
	{ "-movesPerSession", "movesPerSession", XrmoptionSepArg, NULL },
	{ "-mps", "movesPerSession", XrmoptionSepArg, NULL },
	{ "-firstChessProgram", "firstChessProgram", XrmoptionSepArg, NULL },
	{ "-fcp", "firstChessProgram", XrmoptionSepArg, NULL },
	{ "-secondChessProgram", "secondChessProgram", XrmoptionSepArg, NULL },
	{ "-scp", "secondChessProgram", XrmoptionSepArg, NULL },
	{ "-noChessProgram", "noChessProgram", XrmoptionSepArg, NULL },
	{ "-ncp", "noChessProgram", XrmoptionSepArg, NULL },
	{ "-firstHost", "firstHost", XrmoptionSepArg, NULL },
	{ "-fh", "firstHost", XrmoptionSepArg, NULL },
	{ "-secondHost", "secondHost", XrmoptionSepArg, NULL },
	{ "-sh", "secondHost", XrmoptionSepArg, NULL },
	{ "-solidPawnBitmap", "solidPawnBitmap", XrmoptionSepArg, NULL },
	{ "-spb", "solidPawnBitmap", XrmoptionSepArg, NULL },
	{ "-solidRookBitmap", "solidRookBitmap", XrmoptionSepArg, NULL },
	{ "-srb", "solidRookBitmap", XrmoptionSepArg, NULL },
	{ "-solidBishopBitmap", "solidBishopBitmap", XrmoptionSepArg, NULL },
	{ "-sbb", "solidBishopBitmap", XrmoptionSepArg, NULL },
	{ "-solidKnightBitmap", "solidKnightBitmap", XrmoptionSepArg, NULL },
	{ "-snb", "solidKnightBitmap", XrmoptionSepArg, NULL },
	{ "-solidQueenBitmap", "solidQueenBitmap", XrmoptionSepArg, NULL },
	{ "-sqb", "solidQueenBitmap", XrmoptionSepArg, NULL },
	{ "-solidKingBitmap", "solidKingBitmap", XrmoptionSepArg, NULL },
	{ "-skb", "solidKingBitmap", XrmoptionSepArg, NULL },
	{ "-outlinePawnBitmap", "outlinePawnBitmap", XrmoptionSepArg, NULL },
	{ "-opb", "outlinePawnBitmap", XrmoptionSepArg, NULL },
	{ "-outlineRookBitmap", "outlineRookBitmap", XrmoptionSepArg, NULL },
	{ "-orb", "outlineRookBitmap", XrmoptionSepArg, NULL },
	{ "-outlineBishopBitmap", "outlineBishopBitmap",
	    XrmoptionSepArg, NULL },
	{ "-obb", "outlineBishopBitmap", XrmoptionSepArg, NULL },
	{ "-outlineKnightBitmap", "outlineKnightBitmap",
	    XrmoptionSepArg, NULL },
	{ "-onb", "outlineKnightBitmap", XrmoptionSepArg, NULL },
	{ "-outlineQueenBitmap", "outlineQueenBitmap", XrmoptionSepArg, NULL },
	{ "-oqb", "outlineQueenBitmap", XrmoptionSepArg, NULL },
	{ "-outlineKingBitmap", "outlineKingBitmap", XrmoptionSepArg, NULL },
	{ "-okb", "outlineKingBitmap", XrmoptionSepArg, NULL },
	{ "-remoteShell", "remoteShell", XrmoptionSepArg, NULL },
	{ "-rsh", "remoteShell", XrmoptionSepArg, NULL },
	{ "-timeDelay", "timeDelay", XrmoptionSepArg, NULL },
	{ "-td", "timeDelay", XrmoptionSepArg, NULL },
	{ "-timeControl", "timeControl", XrmoptionSepArg, NULL },
	{ "-tc", "timeControl", XrmoptionSepArg, NULL },
	{ "-loadGameFile", "loadGameFile", XrmoptionSepArg, NULL },
	{ "-lgf", "loadGameFile", XrmoptionSepArg, NULL },
	{ "-loadPositionFile", "loadPositionFile", XrmoptionSepArg, NULL },
	{ "-lpf", "loadPositionFile", XrmoptionSepArg, NULL },
	{ "-saveGameFile", "saveGameFile", XrmoptionSepArg, NULL },
	{ "-sgf", "saveGameFile", XrmoptionSepArg, NULL },
	{ "-savePositionFile", "savePositionFile", XrmoptionSepArg, NULL },
	{ "-spf", "savePositionFile", XrmoptionSepArg, NULL },
	{ "-matchMode", "matchMode", XrmoptionSepArg, NULL },
	{ "-mm", "matchMode", XrmoptionSepArg, NULL },
	{ "-monoMode", "monoMode", XrmoptionSepArg, NULL },
	{ "-mono", "monoMode", XrmoptionSepArg, NULL },
	{ "-debugMode", "debugMode", XrmoptionSepArg, NULL },
	{ "-debug", "debugMode", XrmoptionSepArg, NULL },
	{ "-clockMode", "clockMode", XrmoptionSepArg, NULL },
	{ "-clock", "clockMode", XrmoptionSepArg, NULL },
	{ "-boardSize", "boardSize", XrmoptionSepArg, NULL },
	{ "-size", "boardSize", XrmoptionSepArg, NULL },
	{ "-searchTime", "searchTime", XrmoptionSepArg, NULL },
	{ "-st", "searchTime", XrmoptionSepArg, NULL },
	{ "-searchLevel", "searchLevel", XrmoptionSepArg, NULL },
	{ "-sl", "searchLevel", XrmoptionSepArg, NULL },
	{ "-iconic", "Iconic", XrmoptionNoArg, "True" }
};

XtActionsRec boardActions[] = {
	{ "DrawPosition", DrawPosition },
	{ "HandleUserMove", HandleUserMove },
	{ "ResetProc", ResetProc },
	{ "ResetFileProc", ResetFileProc },
	{ "GameProc", GameProc },
	{ "QuitProc", QuitProc },
	{ "ForwardProc", ForwardProc },
	{ "BackwardProc", BackwardProc },
	{ "PauseProc", PauseProc },
	{ "Iconify", Iconify },
	{ "FileNameAction", FileNameAction },
	{ "PieceMenuPopup", PieceMenuPopup },
	{ "SetWhiteToPlay", SetWhiteToPlay },
	{ "SetBlackToPlay", SetBlackToPlay }
};

char translationsTable[] =
	"<Expose>: DrawPosition() \n \
	 <Btn1Down>: HandleUserMove() \n \
	 <Btn1Up>: HandleUserMove() \n \
	 <Btn2Down>: XawPositionSimpleMenu(menuW) PieceMenuPopup(menuW) \n \
	 <Btn3Down>: XawPositionSimpleMenu(menuB) PieceMenuPopup(menuB) \n \
	 <Key>r: ResetFileProc() ResetProc() \n \
	 <Key>R: ResetFileProc() ResetProc() \n \
	 <Key>g: GameProc() \n \
	 <Key>G: GameProc() \n \
	 <Key>q: QuitProc() \n \
	 <Key>Q: QuitProc() \n \
	 <Message>WM_PROTOCOLS: QuitProc() \n \
	 <Key>f: ForwardProc() \n \
	 <Key>F: ForwardProc() \n \
	 <Key>b: BackwardProc() \n \
	 <Key>B: BackwardProc() \n \
	 <Key>p: PauseProc() \n \
	 <Key>P: PauseProc() \n \
	 <Key>i: Iconify() \n \
	 <Key>I: Iconify() \n \
	 <Key>c: Iconify() \n \
	 <Key>C: Iconify() \n";

char whiteTranslations[] = "<BtnDown>: SetWhiteToPlay()\n";
char blackTranslations[] = "<BtnDown>: SetBlackToPlay()\n";

String xboardResources[] = {
	"*font: -*-helvetica-medium-o-normal--*-140-*-*-*-*-*-*\n",
	"*Dialog*value.translations: #override \\n <Key>Return: FileNameAction()",
	NULL
};

void
main(argc, argv)
	int argc;
	char **argv;
{
	XSetWindowAttributes window_attributes;
	char buf[MSG_SIZ];
	Arg args[3];
	Dimension timerWidth, boardWidth, commandsWidth, w;
	extern char *strrchr P((char *, char));

	setbuf(stdout, NULL); setbuf(stderr, NULL);

	programName = strrchr(argv[0], '/');
	if (programName == NULL)
	  programName = argv[0];
	else
	  programName++;

	shellWidget = XtAppInitialize(&appContext, "XBoard", shellOptions,
		XtNumber(shellOptions), &argc, argv, xboardResources, NULL, 0);
	if (argc > 1)
		Usage();

	if ((chessDir = (char *) getenv("CHESSDIR")) == NULL) {
		chessDir = ".";
	} else {
		if (chdir(chessDir) != 0) {
			fprintf(stderr, "%s: can't cd to CHESSDIR\n",
				programName);
			perror(chessDir);
			exit(1);
		}
	}

	XtGetApplicationResources(shellWidget, &appData, clientResources,
		XtNumber(clientResources), NULL, 0);

	xboardDebug = appData.debugMode;

	/*
	 * Determine matchMode state -- poor man's resource converter
	 */
	if (StrCaseCmp(appData.matchMode, "Init") == 0)
		matchMode = MatchInit;
	else if (StrCaseCmp(appData.matchMode, "Position") == 0)
		matchMode = MatchPosition;
	else if (StrCaseCmp(appData.matchMode, "Opening") == 0)
		matchMode = MatchOpening;
	else if (StrCaseCmp(appData.matchMode, "False") == 0)
		matchMode = MatchFalse;
	else {
		fprintf(stderr, "%s: bad matchMode option %s\n",
			appData.matchMode, programName);
		Usage();
	}

	/*
	 * Parse timeControl resource
	 */
	if (appData.timeControl != NULL) {
		int min, sec, matched;

		matched = sscanf(appData.timeControl, "%d:%d", &min, &sec);
		if (matched == 1) {
			timeControl = min * 60;
		} else if (matched == 2) {
			timeControl = min * 60 + sec;
		} else {
			fprintf(stderr, "%s: bad timeControl option %s\n",
				programName, appData.timeControl);
			Usage();
		}
	}

	/*
	 * Determine boardSize
	 */
	if (StrCaseCmp(appData.boardSize, "Large") == 0)
		boardSize = Large;
	else if (StrCaseCmp(appData.boardSize, "Medium") == 0)
		boardSize = Medium;
	else if (StrCaseCmp(appData.boardSize, "Small") == 0)
		boardSize = Small;
	else {
		fprintf(stderr, "%s: bad boardSize option %s\n",
			appData.boardSize, programName);
		Usage();
	}

	xDisplay = XtDisplay(shellWidget);
	xScreen = DefaultScreen(xDisplay);
	if (((DisplayWidth(xDisplay, xScreen) < 800)
	     ||	(DisplayHeight(xDisplay, xScreen) < 800))
	    && (boardSize == Large)) {
		boardSize = Medium;
	}

	switch (boardSize) {
	      case Small:
		squareSize = SMALL_SQUARE_SIZE;
		break;
	      case Medium:
		squareSize = MEDIUM_SQUARE_SIZE;
		break;
	      case Large:
		squareSize = LARGE_SQUARE_SIZE;
		break;
	}
	boardWidth = LINE_GAP + BOARD_SIZE * (squareSize + LINE_GAP);
	XtSetArg(boardArgs[1], XtNwidth, boardWidth);
	XtSetArg(boardArgs[2], XtNheight,
		 LINE_GAP + BOARD_SIZE * (squareSize + LINE_GAP));

        if ((appData.searchTime != NULL) || (appData.searchLevel > 0)
	    || appData.noChessProgram)
	  appData.clockMode = False;

	/*
	 * Detect if there are not enough colors are available and adapt.
	 */
	if (DefaultDepth(xDisplay, xScreen) <= 2)
		appData.monoMode = True;

	/*
	 * widget hierarchy
	 */
	formWidget = XtCreateManagedWidget("form",
		formWidgetClass, shellWidget, NULL, 0);

		widgetList[0] = whiteTimerWidget =
		  XtCreateWidget("white time:", labelWidgetClass,
				 formWidget, timerArgs, XtNumber(timerArgs));

		widgetList[1] = blackTimerWidget =
		  XtCreateWidget("black time:", labelWidgetClass,
				 formWidget, timerArgs, XtNumber(timerArgs));

		widgetList[2] = nameWidget =
		  XtCreateWidget("", labelWidgetClass,
				 formWidget, nameArgs, XtNumber(nameArgs));

		widgetList[3] = messageWidget =
		  XtCreateWidget("message", labelWidgetClass, formWidget,
				 messageArgs, XtNumber(messageArgs));

		widgetList[4] = commandsWidget =
		  XtCreateWidget("commands", listWidgetClass, formWidget,
				 commandsArgs, XtNumber(commandsArgs));

	        widgetList[5] = boardWidget =
		  XtCreateWidget("board", widgetClass, formWidget,
				 boardArgs, XtNumber(boardArgs));

	XtManageChildren(widgetList, XtNumber(widgetList));

	/*
	 * Calculate the width of the timer labels.
	 */
	XtSetArg(args[0], XtNfont, &labelFont);
	XtGetValues(whiteTimerWidget, args, 1);
	if (appData.clockMode) {
		sprintf(buf, "White: %s ", TimeString(timeControl));
		timerWidth = XTextWidth(labelFont, buf, strlen(buf) - 1);
	} else
		timerWidth = XTextWidth(labelFont, "White  ", 7);
	XtSetArg(args[0], XtNwidth, timerWidth);
	XtSetValues(whiteTimerWidget, args, 1);
	XtSetValues(blackTimerWidget, args, 1);

	XtSetArg(args[0], XtNbackground, &timerForegroundPixel);
	XtSetArg(args[1], XtNforeground, &timerBackgroundPixel);
	XtGetValues(whiteTimerWidget, args, 2);

	/*
	 * Calculate the width of the name and message labels.
	 */
	XtSetArg(args[0], XtNwidth, &commandsWidth);
	XtGetValues(commandsWidget, args, 1);
        w = (commandsWidth > boardWidth) ? commandsWidth : boardWidth;
	XtSetArg(args[0], XtNwidth, w - timerWidth*2 - 12);
	XtSetValues(nameWidget, args, 1);
	XtSetArg(args[0], XtNwidth, w - 8);
	XtSetValues(messageWidget, args, 1);

	/*
	 * formWidget uses these constraints but they are stored
	 * in the children.
	 */
	XtSetArg(args[0], XtNfromHoriz, whiteTimerWidget);
	XtSetValues(blackTimerWidget, args, 1);
	XtSetArg(args[0], XtNfromHoriz, blackTimerWidget);
	XtSetValues(nameWidget, args, 1);
	XtSetArg(args[0], XtNfromVert, whiteTimerWidget);
	XtSetValues(messageWidget, args, 1);
	XtSetArg(args[0], XtNfromVert, messageWidget);
	XtSetValues(commandsWidget, args, 1);
	XtSetArg(args[0], XtNfromVert, commandsWidget);
	XtSetValues(boardWidget, args, 1);

	XtRealizeWidget(shellWidget);

	xBoardWindow = XtWindow(boardWidget);

	/*
	 * Create an icon.
	 */
	iconPixmap = XCreateBitmapFromData(xDisplay, XtWindow(shellWidget),
		icon_bits, icon_width, icon_height);
	XtSetArg(args[0], XtNiconPixmap, iconPixmap);
	XtSetValues(shellWidget, args, 1);

	/*
	 * Create a cursor for the board widget.
	 */
	window_attributes.cursor = XCreateFontCursor(xDisplay, XC_hand2);
	XChangeWindowAttributes(xDisplay, xBoardWindow,
		CWCursor, &window_attributes);

	/*
	 * Inhibit shell resizing.
	 */
	XtGetValues(shellWidget, shellArgs, 2);
	shellArgs[4].value = shellArgs[2].value = shellArgs[0].value;
	shellArgs[5].value = shellArgs[3].value = shellArgs[1].value;
	XtSetValues(shellWidget, &shellArgs[2], 4);

	CreateGCs();
	CreateGrid();
	CreatePieces();
	CreatePieceMenus();

	XtAddCallback(commandsWidget, XtNcallback, SelectCommand, NULL);
	XtAppAddActions(appContext, boardActions, XtNumber(boardActions));

	XtSetArg(args[0], XtNtranslations,
		XtParseTranslationTable(translationsTable));
	XtSetValues(boardWidget, &args[0], 1);
	XtSetArg(args[0], XtNtranslations,
		XtParseTranslationTable(whiteTranslations));
	XtSetValues(whiteTimerWidget, &args[0], 1);
	XtSetArg(args[0], XtNtranslations,
		XtParseTranslationTable(blackTranslations));
	XtSetValues(blackTimerWidget, &args[0], 1);

	XtAddEventHandler(boardWidget, ExposureMask | ButtonPressMask
		| ButtonReleaseMask | Button1MotionMask | KeyPressMask,
		False, EventProc, NULL);

        sprintf(buf, "xboard version %s, patchlevel %d", VERSION, PATCHLEVEL);

	/*
	 * If there is to be a machine match, set it up.
	 */
	if (matchMode != MatchFalse){
		if (appData.noChessProgram) {
			fprintf(stderr,
			  "%s: can't have a match with no chess programs!\n",
			  programName);
			exit(1);
		}
		DisplayMessage(buf);
		TwoMachinesPlayProc();
	}
	else {
		ResetProc();
		DisplayMessage(buf);
	}

	XtAppMainLoop(appContext);
}

void
CreateGCs()
{
	XtGCMask value_mask = GCLineWidth | GCLineStyle | GCForeground
		| GCBackground | GCFunction | GCPlaneMask;
	XGCValues gc_values;

	gc_values.plane_mask = AllPlanes;
	gc_values.line_width = LINE_GAP;
	gc_values.line_style = LineSolid;
	gc_values.function = GXcopy;

	gc_values.foreground = XBlackPixel(xDisplay, xScreen);
	gc_values.background = XBlackPixel(xDisplay, xScreen);
	lineGC = XtGetGC(shellWidget, value_mask, &gc_values);

	if (appData.monoMode) {
		gc_values.foreground = XWhitePixel(xDisplay, xScreen);
		gc_values.background = XBlackPixel(xDisplay, xScreen);
		lightSquareGC = wbPieceGC
			= XtGetGC(shellWidget, value_mask, &gc_values);

		gc_values.foreground = XBlackPixel(xDisplay, xScreen);
		gc_values.background = XWhitePixel(xDisplay, xScreen);
		darkSquareGC = bwPieceGC
			= XtGetGC(shellWidget, value_mask, &gc_values);
	} else {
		gc_values.foreground = appData.lightSquareColor;
		gc_values.background = appData.darkSquareColor;
		lightSquareGC = XtGetGC(shellWidget, value_mask, &gc_values);

		gc_values.foreground = appData.darkSquareColor;
		gc_values.background = appData.lightSquareColor;
		darkSquareGC = XtGetGC(shellWidget, value_mask, &gc_values);

		gc_values.foreground = appData.whitePieceColor;
		gc_values.background = appData.darkSquareColor;
		wdPieceGC = XtGetGC(shellWidget, value_mask, &gc_values);

		gc_values.foreground = appData.whitePieceColor;
		gc_values.background = appData.lightSquareColor;
		wlPieceGC = XtGetGC(shellWidget, value_mask, &gc_values);

		gc_values.foreground = appData.blackPieceColor;
		gc_values.background = appData.darkSquareColor;
		bdPieceGC = XtGetGC(shellWidget, value_mask, &gc_values);

		gc_values.foreground = appData.blackPieceColor;
		gc_values.background = appData.lightSquareColor;
		blPieceGC = XtGetGC(shellWidget, value_mask, &gc_values);
	}
}

void
CreatePieces()
{
	XSynchronize(xDisplay, True);	/* Work-around for xlib/xt
					   buffering bug */

	ReadBitmap(appData.solidPawnBitmap, &solidPawnBitmap,
		   solid_pawn_bits, pawn_small_bits, xs_s_p_bits);
	ReadBitmap(appData.solidRookBitmap, &solidRookBitmap,
		   solid_rook_bits, rook_small_bits, xs_s_r_bits);
	ReadBitmap(appData.solidKnightBitmap, &solidKnightBitmap,
		   solid_knight_bits, knight_small_bits, xs_s_kt_bits);
	ReadBitmap(appData.solidBishopBitmap, &solidBishopBitmap,
		   solid_bishop_bits, bishop_small_bits, xs_s_b_bits);
	ReadBitmap(appData.solidQueenBitmap, &solidQueenBitmap,
		   solid_queen_bits, queen_small_bits, xs_s_q_bits);
	ReadBitmap(appData.solidKingBitmap, &solidKingBitmap,
		   solid_king_bits, king_small_bits, xs_s_k_bits);

	if (appData.monoMode) {
		ReadBitmap(appData.outlinePawnBitmap, &outlinePawnBitmap,
			   outline_pawn_bits, pawn_small_outline_bits,
			   xs_ol_p_bits);
		ReadBitmap(appData.outlineRookBitmap, &outlineRookBitmap,
			   outline_rook_bits, rook_small_outline_bits,
			   xs_ol_r_bits);
		ReadBitmap(appData.outlineKnightBitmap, &outlineKnightBitmap,
			   outline_knight_bits, knight_small_outline_bits,
			   xs_ol_kt_bits);
		ReadBitmap(appData.outlineBishopBitmap, &outlineBishopBitmap,
			   outline_bishop_bits, bishop_small_outline_bits,
			   xs_ol_b_bits);
		ReadBitmap(appData.outlineQueenBitmap, &outlineQueenBitmap,
			   outline_queen_bits, queen_small_outline_bits,
			   xs_ol_q_bits);
		ReadBitmap(appData.outlineKingBitmap, &outlineKingBitmap,
			   outline_king_bits, king_small_outline_bits,
			   xs_ol_k_bits);
	}

	XSynchronize(xDisplay, False);	/* Work-around for xlib/xt
					   buffering bug */
}

void
ReadBitmap(name, pm, large_bits, medium_bits, small_bits)
	String name;
	Pixmap *pm;
	char large_bits[], medium_bits[], small_bits[];
{
	int x_hot, y_hot;
	u_int w, h;

	if (name == NULL || XReadBitmapFile(xDisplay, xBoardWindow, name,
		&w, &h, pm, &x_hot, &y_hot) != BitmapSuccess
		|| w != squareSize || h != squareSize) {
		switch (boardSize) {
		      case Large:
			*pm = XCreateBitmapFromData(xDisplay, xBoardWindow,
				large_bits, squareSize, squareSize);
			break;
		      case Medium:
			*pm = XCreateBitmapFromData(xDisplay, xBoardWindow,
				medium_bits, squareSize, squareSize);
			break;
		      case Small:
			*pm = XCreateBitmapFromData(xDisplay, xBoardWindow,
				small_bits, squareSize, squareSize);
			break;
		}
	}
}

void
CreateGrid()
{
	int i;

	for (i = 0; i < BOARD_SIZE + 1; i++) {
		gridSegments[i].x1 = gridSegments[i + BOARD_SIZE + 1].y1 = 0;
		gridSegments[i].y1 = gridSegments[i].y2
			= LINE_GAP / 2 + (i * (squareSize + LINE_GAP));
		gridSegments[i].x2 = LINE_GAP + BOARD_SIZE *
		        (squareSize + LINE_GAP);
		gridSegments[i + BOARD_SIZE + 1].x1 =
		gridSegments[i + BOARD_SIZE + 1].x2 = LINE_GAP / 2
			+ (i * (squareSize + LINE_GAP));
		gridSegments[i + BOARD_SIZE + 1].y2 =
			BOARD_SIZE * (squareSize + LINE_GAP);
	}
}

void
CreatePieceMenus()
{
	int i;
	Widget entry;
	Arg args[1];
	ChessSquare selection;
	
	XtSetArg(args[0], XtNlabel, "White");
	whitePieceMenu = XtCreatePopupShell("menuW", simpleMenuWidgetClass,
				       boardWidget, args, 1);
	for (i = 0; i < PIECE_MENU_SIZE; i++) {
		String item = pieceMenuStrings[i];

		if (strcmp(item, "----") == 0) {
			entry = XtCreateManagedWidget(item, smeLineObjectClass,
						      whitePieceMenu, NULL, 0);
		} else {
			entry = XtCreateManagedWidget(item, smeBSBObjectClass,
						      whitePieceMenu, NULL, 0);
			selection = pieceMenuTranslation[0][i];
			XtAddCallback(entry, XtNcallback, PieceMenuSelect,
				      (caddr_t) selection);
		}
		if (selection == WhitePawn) {
			XtSetArg(args[0], XtNpopupOnEntry, entry);
			XtSetValues(whitePieceMenu, args, 1);
		}
	}

	XtSetArg(args[0], XtNlabel, "Black");
	blackPieceMenu = XtCreatePopupShell("menuB", simpleMenuWidgetClass,
				       boardWidget, args, 1);
	for (i = 0; i < PIECE_MENU_SIZE; i++) {
		String item = pieceMenuStrings[i];

		if (strcmp(item, "----") == 0) {
			entry = XtCreateManagedWidget(item, smeLineObjectClass,
						      blackPieceMenu, NULL, 0);
		} else {
			entry = XtCreateManagedWidget(item, smeBSBObjectClass,
						      blackPieceMenu, NULL, 0);
			selection = pieceMenuTranslation[1][i];
			XtAddCallback(entry, XtNcallback, PieceMenuSelect,
				      (caddr_t) selection);
		}
		if (selection == BlackPawn) {
			XtSetArg(args[0], XtNpopupOnEntry, entry);
			XtSetValues(blackPieceMenu, args, 1);
		}
	}

	XtRegisterGrabAction(PieceMenuPopup, True,
			     (unsigned)(ButtonPressMask|ButtonReleaseMask),
			     GrabModeAsync, GrabModeAsync);
}	

void
PieceMenuPopup(w, event, params, num_params)
	Widget w;
	XEvent *event;
        String *params;
        Cardinal *num_params;
{
	if (event->type != ButtonPress) return;
	if (gameMode != EditPosition) return;

	if (((pmFromX = EventToSquare(event->xbutton.x)) < 0) ||
	    ((pmFromY = EventToSquare(event->xbutton.y)) < 0)) {
		pmFromX = pmFromY = -1;
		return;
	}
	if (flipView)
	  pmFromX = BOARD_SIZE - 1 - pmFromX;
	else
	  pmFromY = BOARD_SIZE - 1 - pmFromY;

	XtPopupSpringLoaded(XtNameToWidget(boardWidget, params[0]));
}

static void
PieceMenuSelect(w, piece, junk)
     Widget w;
     ChessSquare piece;
     caddr_t junk;
{
	if (pmFromX < 0 || pmFromY < 0) return;
	switch (piece) {
	      case ClearBoard:
		for (pmFromY = 0; pmFromY < BOARD_SIZE; pmFromY++)
		  for (pmFromX = 0; pmFromX < BOARD_SIZE; pmFromX++) {
			  boards[0][pmFromY][pmFromX] = EmptySquare;
			  DrawSquare(pmFromY, pmFromX, EmptySquare);
		  }
		break;

	      case WhitePlay:  /*not currently on menu*/
		SetWhiteToPlay();
		break;

	      case BlackPlay:  /*not currently on menu*/
		SetBlackToPlay();
		break;

	      default:
		boards[0][pmFromY][pmFromX] = piece;
		DrawSquare(pmFromY, pmFromX, piece);
		break;
	}
	XSync(xDisplay, False);
}

static void
SetWhiteToPlay()
{
	int saveCM;

	if (gameMode != EditPosition) return;
	blackPlaysFirst = False;
	saveCM = currentMove;
	currentMove = 0;  /*kludge*/
	DisplayClocks(ReDisplayTimers);
	currentMove = saveCM;
}

static void
SetBlackToPlay()
{
	int saveCM;

	if (gameMode != EditPosition) return;
	blackPlaysFirst = True;
	saveCM = currentMove;
	currentMove = 1;  /*kludge*/
	DisplayClocks(ReDisplayTimers);
	currentMove = saveCM;
}

/*
 * If the user selects on a border boundary or off the board, return failure.
 * Otherwise map the event coordinate to the square.
 */
int
EventToSquare(x)
	int x;
{
	if (x < LINE_GAP)
		return -1;
	x -= LINE_GAP;
	if ((x % (squareSize + LINE_GAP)) >= squareSize)
		return -1;
	x /= (squareSize + LINE_GAP);
	if (x >= BOARD_SIZE)
		return -1;
	return x;
}

ChessSquare
CharToPiece(c)
	int c;
{
	switch (c) {
	default:
	case '.':	return EmptySquare;
	case 'P':	return WhitePawn;
	case 'R':	return WhiteRook;
	case 'N':	return WhiteKnight;
	case 'B':	return WhiteBishop;
	case 'Q':	return WhiteQueen;
	case 'K':	return WhiteKing;
	case 'p':	return BlackPawn;
	case 'r':	return BlackRook;
	case 'n':	return BlackKnight;
	case 'b':	return BlackBishop;
	case 'q':	return BlackQueen;
	case 'k':	return BlackKing;
	}
}

void
DrawSquare(row, column, piece)
	int row, column;
	ChessSquare piece;
{
	int square_color, x, y;

	if (flipView)
		column = (BOARD_SIZE - 1) - column;
	else
		row = (BOARD_SIZE - 1) - row;

	square_color = ((column % 2 == 1) && (row % 2 == 1))
		|| ((column % 2 == 0) && (row % 2 == 0));

	x = LINE_GAP + column * (squareSize + LINE_GAP);
	y = LINE_GAP + row * (squareSize + LINE_GAP);

	if (piece == EmptySquare)
		XFillRectangle(xDisplay, xBoardWindow,
			       square_color ? lightSquareGC : darkSquareGC,
			       x, y, squareSize, squareSize);
	else if (appData.monoMode) {
		if (square_color)
		  XCopyPlane(xDisplay, (int) piece < (int) BlackPawn
			     ? *pieceToOutline[(int) piece]
			     : *pieceToSolid[(int) piece],
			     xBoardWindow, bwPieceGC, 0, 0,
			     squareSize, squareSize, x, y, 1);
		else
		  XCopyPlane(xDisplay, (int) piece < (int) BlackPawn
			     ? *pieceToSolid[(int) piece]
			     : *pieceToOutline[(int) piece],
			     xBoardWindow, wbPieceGC, 0, 0,
			     squareSize, squareSize, x, y, 1);
	} else {
		if (square_color)
		  XCopyPlane(xDisplay, *pieceToSolid[(int) piece],
			     xBoardWindow, (int) piece < (int) BlackPawn
			     ? wlPieceGC : blPieceGC, 0, 0,
			     squareSize, squareSize, x, y, 1);
		else
		  XCopyPlane(xDisplay, *pieceToSolid[(int) piece],
			     xBoardWindow, (int) piece < (int) BlackPawn
			     ? wdPieceGC : bdPieceGC, 0, 0,
			     squareSize, squareSize, x, y, 1);
	}
}

void
EventProc(widget, unused, event)
	Widget widget;
	caddr_t unused;
	XEvent *event;
{
	if (event->type == MappingNotify) {
		XRefreshKeyboardMapping((XMappingEvent *) event);
		return;
	}

	if (!XtIsRealized(widget))
		return;

	if ((event->type == ButtonPress) || (event->type == ButtonRelease))
		if (event->xbutton.button != Button1)
			return;

	switch (event->type) {
	case Expose:
		DrawPosition(widget, (XExposeEvent *) event);
		break;
	default:
		return;
	}
}

/*
 * event handler for redrawing the board
 */
void
DrawPosition(w, event)
	Widget w;
	XExposeEvent *event;
{
	Arg args[1];
	int i, j;
        if (!appData.Iconic){
		XtSetArg(args[0], XtNiconic, False);
		XtSetValues(shellWidget, args, 1);
        }

	/*
	 * It would be simpler to clear the window with XClearWindow()
	 * but this causes a very distracting flicker.
	 */
	XDrawSegments(xDisplay, xBoardWindow, lineGC,
		gridSegments, (BOARD_SIZE + 1) * 2);

	for (i = 0; i < BOARD_SIZE; i++)
		for (j = 0; j < BOARD_SIZE; j++)
			DrawSquare(i, j, boards[currentMove][i][j]);

	XSync(xDisplay, False);
}

void
InitPosition()
{
	currentMove = forwardMostMove = 0;
	CopyBoard(boards[0], initialPosition);
	DrawPosition(boardWidget, (XExposeEvent *) NULL);
}

void
CopyBoard(to, from)
	Board to, from;
{
	int i, j;

	for (i = 0; i < BOARD_SIZE; i++)
		for (j = 0; j < BOARD_SIZE; j++)
			to[i][j] = from[i][j];
}

void
SendCurrentBoard(fp)
	FILE *fp;
{
	SendBoard(fp, boards[currentMove]);
}

void
SendBoard(fp, board)
	FILE *fp;
        Board board;
{
	char message[MSG_SIZ];
	ChessSquare *bp;
	int i, j;

	SendToProgram("edit\n", fp);
	SendToProgram("#\n", fp);
	for (i = BOARD_SIZE - 1; i >= 0; i--) {
		bp = &board[i][0];
		for (j = 0; j < BOARD_SIZE; j++, bp++) {
			if ((int) *bp < (int) BlackPawn) {
				sprintf(message, "%c%c%c\n",
					pieceToChar[(int) *bp],
					'a' + j, '1' + i);
				SendToProgram(message, fp);
			}
		}
	}

	SendToProgram("c\n", fp);
	for (i = BOARD_SIZE - 1; i >= 0; i--) {
		bp = &board[i][0];
		for (j = 0; j < BOARD_SIZE; j++, bp++) {
			if (((int) *bp != (int) EmptySquare)
				&& ((int) *bp >= (int) BlackPawn)) {
				sprintf(message, "%c%c%c\n",
					pieceToChar[(int) *bp -
						    (int) BlackPawn],
					'a' + j, '1' + i);
				SendToProgram(message, fp);
			}
		}
	}

	SendToProgram(".\n", fp);
}

/*
 * event handler for parsing user moves
 */
void
HandleUserMove(w, event)
	Widget w;
	XEvent *event;
{
	ChessMove move_type;
	ChessSquare from_piece;
	int to_x, to_y;

	if ((w != boardWidget) || (matchMode != MatchFalse))
		return;

	if (promotionUp) {
		XtPopdown(promotionShell);
		XtDestroyWidget(promotionShell);
		promotionUp = False;
		fromX = fromY = -1;
	}

	switch (gameMode) {
	      case EndOfGame:
	      case PlayFromGameFile:
	      case TwoMachinesPlay:
		return;
	      case MachinePlaysWhite:
		if (WHITE_ON_MOVE)
			return;
		break;
	      case MachinePlaysBlack:
		if (!WHITE_ON_MOVE)
			return;
		break;
	      default:
		break;
	}

	switch (event->type) {
	      case ButtonPress:
		if ((fromX >= 0) || (fromY >= 0))
		  return;
		if (((fromX = EventToSquare(event->xbutton.x)) < 0) ||
		    ((fromY = EventToSquare(event->xbutton.y)) < 0)) {
			fromX = fromY = -1;
			return;
		}
		if (flipView)
		  fromX = BOARD_SIZE - 1 - fromX;
		else
		  fromY = BOARD_SIZE - 1 - fromY;
		break;

	      case ButtonRelease:
		if ((fromX < 0) || (fromY < 0)) return;

		if (((to_x = EventToSquare(event->xbutton.x)) < 0)
			|| ((to_y = EventToSquare(event->xbutton.y)) < 0)) {
			if (gameMode == EditPosition) {
				boards[0][fromY][fromX] = EmptySquare;
				DrawSquare(fromY, fromX, EmptySquare);
				XSync(xDisplay, False);
			}
			fromX = fromY = -1;
			return;
		}
		if (flipView)
			to_x = BOARD_SIZE - 1 - to_x;
		else
			to_y = BOARD_SIZE - 1 - to_y;

		if ((fromX == to_x) && (fromY == to_y)) {
			fromX = fromY = -1;
			return;
		}

		if (gameMode == EditPosition) {
			boards[0][to_y][to_x] = boards[0][fromY][fromX];
			DrawSquare(to_y, to_x, boards[0][to_y][to_x]);
			boards[0][fromY][fromX] = EmptySquare;
			DrawSquare(fromY, fromX, EmptySquare);
			fromX = fromY = -1;
			XSync(xDisplay, False);
			return;
		}
		
		from_piece = boards[currentMove][fromY][fromX];
		if (((from_piece == WhitePawn && to_y == 7) ||
		     (from_piece == BlackPawn && to_y == 0))) {
			PromotionPopUp(from_piece, to_x, to_y);
			return;
		}
		move_type = NormalMove;
		MakeMove(&move_type, fromX, fromY, to_x, to_y);
		FinishUserMove(move_type, to_x, to_y);
		break;
	}
}

void
FinishUserMove(move_type, to_x, to_y)
         ChessMove move_type;
         int to_x, to_y;
{
	char user_move[MSG_SIZ];

	switch (move_type) {
	      default:
		fprintf(stderr, "%s: internal error; bad move_type\n",
			programName);
		break;
	      case WhitePromotionQueen:
	      case BlackPromotionQueen:
		sprintf(user_move, "%c%c%c%cq\n",
			'a' + fromX, '1' + fromY, 'a' + to_x, '1' + to_y);
		break;
	      case WhitePromotionRook:
	      case BlackPromotionRook:
		sprintf(user_move, "%c%c%c%cr\n",
			'a' + fromX, '1' + fromY, 'a' + to_x, '1' + to_y);
		break;
	      case WhitePromotionBishop:
	      case BlackPromotionBishop:
		sprintf(user_move, "%c%c%c%cb\n",
			'a' + fromX, '1' + fromY, 'a' + to_x, '1' + to_y);
		break;
	      case WhitePromotionKnight:
	      case BlackPromotionKnight:
		sprintf(user_move, "%c%c%c%cn\n",
			'a' + fromX, '1' + fromY, 'a' + to_x, '1' + to_y);
		break;
	      case NormalMove:
	      case WhiteKingSideCastle:
	      case WhiteQueenSideCastle:
	      case WhiteCapturesEnPassant:
	      case BlackKingSideCastle:
	      case BlackQueenSideCastle:
	      case BlackCapturesEnPassant:
		sprintf(user_move, "%c%c%c%c\n",
			'a' + fromX, '1' + fromY, 'a' + to_x, '1' + to_y);
		break;
	}

	fromX = fromY = -1;

	Attention(firstProgramPID);
	SendToProgram(user_move, toFirstProgFP);
	strcpy(moveList[currentMove - 1], user_move);

	if (gameMode == PauseGame)
	  PauseProc(); /* a user move restarts a paused game*/

	switch (gameMode) {
	      case ForceMoves:
	      case SetupPosition:
		break;
	      case BeginningOfGame:
		if (appData.noChessProgram)
		  lastGameMode = gameMode = ForceMoves;
		else
		  lastGameMode = gameMode = MachinePlaysBlack;
		ModeHighlight();
		break;
	      case MachinePlaysBlack:
	      case MachinePlaysWhite:
	      default:
		break;
	}
}

void
HandleMachineMove(message, fp)
	char *message;
        FILE *fp;
{
	char machine_move[MSG_SIZ], buf1[MSG_SIZ], buf2[MSG_SIZ];
	int i, j, from_x, from_y, to_x, to_y;
	ChessMove move_type;

	if (strncmp(message, "warning:", 8) == 0) {
		DisplayMessage(message);
		return;
	}

	/*
	 * If chess program startup fails, exit with an error message.
	 * Attempts to recover here are futile.
	 */
	if ((StrStr(message, "unknown host") != NULL)
	    || (StrStr(message, "No remote directory") != NULL)
	    || (StrStr(message, "not found") != NULL)
	    || (StrStr(message, "No such file") != NULL)
	    || (StrStr(message, "Permission denied") != NULL)) {
		fprintf(stderr,
			"%s: failed to start chess program %s on %s: %s\n",
			programName,
			fp == fromFirstProgFP ? appData.firstChessProgram
			    : appData.secondChessProgram,
			fp == fromFirstProgFP ? appData.firstHost
			    : appData.secondHost,
			message);
		ShutdownChessPrograms(message);
		exit(1);
	}

	/*
	 * If the move is illegal, cancel it and redraw the board.
	 */
	if (strncmp(message, "Illegal move", 12) == 0) {

		if (currentMove <= (blackPlaysFirst ? 1 : 0)) return;
		if (gameMode == PlayFromGameFile) {
			/* Stop reading this game file */
			gameMode = ForceMoves;
			ModeHighlight();
		}
		currentMove--;
		if ((gameMode == PlayFromGameFile) || 
		    (gameMode == ForceMoves) ||
		    (gameMode == SetupPosition))
		  DisplayClocks(ReDisplayTimers);
		else
		  DisplayClocks(DisplayTimers);
		sprintf(buf1, "Illegal move: %s", moveList[currentMove]);
		DisplayMessage(buf1);

		/*
		 * Only redraw the squares that have changed.
		 */
		for (i = 0; i < BOARD_SIZE; i++)
		  for (j = 0; j < BOARD_SIZE; j++) {
			  if (boards[currentMove][i][j] !=
			      boards[currentMove + 1][i][j])
			    DrawSquare(i, j, boards[currentMove][i][j]);
		  }

		XSync(xDisplay, False);
		return;
	}

	if (strncmp(message, "Hint:", 5) == 0) {
		DisplayMessage(message);
		return;
	}

	/*
	 * win, lose or draw
	 */
	if (strncmp(message, "White", 5) == 0) {
		ShutdownChessPrograms("White wins");
		return;
	} else if (strncmp(message, "Black", 5) == 0) {
		ShutdownChessPrograms("Black wins");
		return;
	} else if (strncmp(message, "opponent mates!", 15) == 0) {
		switch (gameMode) {
		      case MachinePlaysBlack:
			ShutdownChessPrograms("White wins");
			break;
		      case MachinePlaysWhite:
			ShutdownChessPrograms("Black wins");
			break;
		      case TwoMachinesPlay:
			ShutdownChessPrograms(fp == fromFirstProgFP ?
					      "White wins" : "Black wins");
			break;
		}
		return;
	} else if (strncmp(message, "computer mates!", 15) == 0) {
		switch (gameMode) {
		      case MachinePlaysBlack:
			ShutdownChessPrograms("Black wins");
			break;
		      case MachinePlaysWhite:
			ShutdownChessPrograms("White wins");
			break;
		      case TwoMachinesPlay:
			ShutdownChessPrograms(fp == fromFirstProgFP ?
					      "Black wins" : "White wins");
			break;
		}
		return;
	} else if (strncmp(message, "Draw", 4) == 0) {
		ShutdownChessPrograms("Draw");
		return;
	}

	/*
	 * normal machine reply move
	 */
	if (StrStr(message, "...") != NULL) {
		sscanf(message, "%s %s %s", buf1, buf2, machine_move);
		if (machine_move[0] == '\0')
			return;
	} else
		return; /* ignore noise */

	strcpy(moveList[currentMove], machine_move);

	from_x = machine_move[0] - 'a';
	from_y = machine_move[1] - '1';
	to_x = machine_move[2] - 'a';
	to_y = machine_move[3] - '1';

	switch (machine_move[4]) {
	      case 'q':
		move_type = WHITE_ON_MOVE ?
		  WhitePromotionQueen : BlackPromotionQueen;
		break;
	      case 'r':
		move_type = WHITE_ON_MOVE ?
		  WhitePromotionRook : BlackPromotionRook;
		break;
	      case 'b':
		move_type = WHITE_ON_MOVE ?
		  WhitePromotionBishop : BlackPromotionBishop;
		break;
	      case 'n':
		move_type = WHITE_ON_MOVE ?
		  WhitePromotionKnight : BlackPromotionKnight;
		break;
	      default:
		move_type = NormalMove;
		break;
	}		  

	MakeMove(&move_type, from_x, from_y, to_x, to_y);

	switch (gameMode) {
	      case PauseGame:
	      case EndOfGame:
	      default:
		break;
	      case ForceMoves:
	      case SetupPosition:
	      case PlayFromGameFile:
		strncat(machine_move, "\n", 1);
		SendToProgram(machine_move, toFirstProgFP);
		break;
	      case TwoMachinesPlay:
		strncat(machine_move, "\n", 1);
		if (WHITE_ON_MOVE) {
			Attention(secondProgramPID);
			SendToProgram(machine_move, toSecondProgFP);
			if (firstMove) {
				firstMove = False;
				SendToProgram(appData.whiteString,
					      toSecondProgFP);
			}
		} else {
			Attention(firstProgramPID);
			SendToProgram(machine_move, toFirstProgFP);
			if (firstMove) {
				firstMove = False;
				SendToProgram(appData.blackString,
					      toFirstProgFP);
			}
		}
		break;
	}
}

void
ReadGameFile()
{
	for (;;) {
		if (!ReadGameFileProc())
			return;
		if (matchMode == MatchOpening)
			continue;
		readGameXID = XtAppAddTimeOut(appContext,
			(int) (1000 * appData.timeDelay), ReadGameFile, NULL);
		break;
	}
}

int
ReadGameFileProc()
{
	int from_x, from_y, to_x, to_y, done;
	ChessMove move_type;
	char move[MSG_SIZ];

	if (gameFileFP == NULL)
		return (int) False;

	if (gameMode == PauseGame) return True;

	if (gameMode != PlayFromGameFile) {
		fclose(gameFileFP);
		gameFileFP = NULL;
		yynewfile();
		return (int) False;
	}

	if (commentUp) {
		XtPopdown(commentShell);
		XtDestroyWidget(commentShell);
		commentUp = False;
	}

	move_type = (ChessMove) yylex();

	if (xboardDebug) {
		switch (move_type) {
		      case BadMove:
			fprintf(stderr, "Parsed BadMove: %s\n", yytext);
			break;
		      case Comment:
			fprintf(stderr, "Parsed Comment: %s\n", yytext);
			break;
		      case WhiteWins:
		      case BlackWins:
		      case GameIsDrawn:
			fprintf(stderr, "Parsed game end: %s\n", yytext);
			break;
		      case StartGame:
			fprintf(stderr, "Parsed StartGame: %s\n", yytext);
			break;
		      default:
			fprintf(stderr, "Parsed %s into %s",
				yytext, currentMoveString);
			break;
		}
	}

	done = False;
	switch (move_type) {
	case Comment:
		CommentPopUp(yytext);
		return (int) True;
	case WhiteKingSideCastle:
		from_x = 4;
		from_y = 0;
		to_x = 6;
		to_y = 0;
		break;
	case WhiteQueenSideCastle:
		from_x = 4;
		from_y = 0;
		to_x = 2;
		to_y = 0;
		break;
	case BlackKingSideCastle:
		from_x = 4;
		from_y = 7;
		to_x = 6;
		to_y = 7;
		break;
	case BlackQueenSideCastle:
		from_x = 4;
		from_y = 7;
		to_x = 2;
		to_y = 7;
		break;
	case WhiteCapturesEnPassant:
	case BlackCapturesEnPassant:
	case WhitePromotionQueen:
	case BlackPromotionQueen:
	case WhitePromotionRook:
	case BlackPromotionRook:
	case WhitePromotionBishop:
	case BlackPromotionBishop:
	case WhitePromotionKnight:
	case BlackPromotionKnight:
	case NormalMove:
		from_x = tolower(currentMoveString[0]) - 'a';
		from_y = tolower(currentMoveString[1]) - '1';
		to_x = tolower(currentMoveString[2]) - 'a';
		to_y = tolower(currentMoveString[3]) - '1';
		break;
	case 0:	/* end of file */
		lastGameMode = gameMode;
		gameMode = ForceMoves;
		ModeHighlight();
		DisplayMessage("End of game file");
		done = True;
		break;
	case WhiteWins:
	case BlackWins:
	case GameIsDrawn:
		lastGameMode = gameMode;
		gameMode = ForceMoves;
		ModeHighlight();
		DisplayMessage(yytext);
		done = True;
		break;
	case StartGame:
		/* ignore (?) */
		return (int) True;
	default:
	case BadMove:
		if (WHITE_ON_MOVE)
		  sprintf(move, "Bad move: %d. %s",
			  (currentMove / 2) + 1, yytext);
		else
		  sprintf(move, "Bad move: %d. ... %s",
			  (currentMove / 2) + 1, yytext);
		DisplayMessage(move);
		done = True;
		break;
	}

	if (done) {
		if (readGameXID != NULL) {
			XtRemoveTimeOut(readGameXID);
			readGameXID = NULL;
		}
		fclose(gameFileFP);
		gameFileFP = NULL;
		yynewfile();
		return (int) False;
	} else {
		SendToProgram(currentMoveString, toFirstProgFP);
		strcpy(moveList[currentMove], currentMoveString);

		MakeMove(&move_type, from_x, from_y, to_x, to_y);

		return (int) True;
	}
}

/*
 * MakeMove() displays moves.  If they are illegal, GNU chess will detect
 * this and send an Illegal move message.  XBoard will then retract the move.
 * The clockMode False case is tricky because it displays the player on move.
 */
void
MakeMove(move_type, from_x, from_y, to_x, to_y)
	ChessMove *move_type;
	int from_x, from_y, to_x, to_y;
{
	char message[MSG_SIZ];
	int move;

	CopyBoard(boards[currentMove + 1], boards[currentMove]);
	forwardMostMove = ++currentMove;

	move = (currentMove + 1) / 2;

	if ((gameMode == PlayFromGameFile) || (gameMode == ForceMoves) ||
	    (gameMode == SetupPosition))
	  DisplayClocks(ReDisplayTimers);
	else
	  DisplayClocks(DisplayTimers);

	if (from_y == 0 && from_x == 4
			&& boards[currentMove][from_y][from_x] == WhiteKing
			&& to_y == 0 && to_x == 6) {
		*move_type = WhiteKingSideCastle;
		boards[currentMove][0][7] = EmptySquare;
		boards[currentMove][0][6] = WhiteKing;
		boards[currentMove][0][5] = WhiteRook;
		boards[currentMove][0][4] = EmptySquare;
		DrawSquare(0, 7, boards[currentMove][0][7]);
		DrawSquare(0, 6, boards[currentMove][0][6]);
		DrawSquare(0, 5, boards[currentMove][0][5]);
		DrawSquare(0, 4, boards[currentMove][0][4]);
		if (gameMode == PlayFromGameFile)
			sprintf(message, "%d. %s", move, yytext);
		else
			sprintf(message, "%d. 0-0", move);
	} else if (from_y == 0 && from_x == 4
			&& boards[currentMove][from_y][from_x] == WhiteKing
			&& to_y == 0 && to_x == 2) {
		*move_type = WhiteQueenSideCastle;
		boards[currentMove][0][0] = EmptySquare;
		boards[currentMove][0][2] = WhiteKing;
		boards[currentMove][0][3] = WhiteRook;
		boards[currentMove][0][4] = EmptySquare;
		DrawSquare(0, 0, boards[currentMove][0][0]);
		DrawSquare(0, 2, boards[currentMove][0][2]);
		DrawSquare(0, 3, boards[currentMove][0][3]);
		DrawSquare(0, 4, boards[currentMove][0][4]);
		if (gameMode == PlayFromGameFile)
			sprintf(message, "%d. %s", move, yytext);
		else
			sprintf(message, "%d. 0-0-0", move);
	} else if (from_y == 6
		   && boards[currentMove][from_y][from_x] == WhitePawn
		   && to_y == 7) {
		/* white pawn promotion */
		char pieceChar;

		switch (*move_type) {
		      case WhitePromotionQueen:
			boards[currentMove][7][to_x] = WhiteQueen;
			pieceChar = 'Q';
			break;
		      case WhitePromotionRook:
        		boards[currentMove][7][to_x] = WhiteRook;
			pieceChar = 'R';
			break;
		      case WhitePromotionBishop:
        		boards[currentMove][7][to_x] = WhiteBishop;
			pieceChar = 'B';
			break;
		      case WhitePromotionKnight:
        		boards[currentMove][7][to_x] = WhiteKnight;
			pieceChar = 'N';
			break;
		      default:
			*move_type = WhitePromotionQueen;
        		boards[currentMove][7][to_x] = WhiteQueen;
			pieceChar = 'Q';
			break;
		}

		boards[currentMove][6][from_x] = EmptySquare;
		DrawSquare(6, from_x, boards[currentMove][6][from_x]);
		DrawSquare(7, to_x, boards[currentMove][7][to_x]);
		if (gameMode == PlayFromGameFile)
			sprintf(message, "%d. %s", move, yytext);
		else
			sprintf(message, "%d. %c7%c8(%c)", 
				move, from_x + 'a', to_x + 'a', pieceChar);
	} else if ((from_y == 4)
			&& (to_x != from_x)
			&& (boards[currentMove][from_y][from_x] == WhitePawn)
			&& (boards[currentMove][to_y][to_x] == EmptySquare)) {
		*move_type = WhiteCapturesEnPassant;
		boards[currentMove][from_y][from_x] = EmptySquare;
		boards[currentMove][to_y][to_x] = WhitePawn;
		boards[currentMove][to_y - 1][to_x] = EmptySquare;
		DrawSquare(from_y, from_x,
			   boards[currentMove][from_y][from_x]);
		DrawSquare(to_y, to_x, boards[currentMove][to_y][to_x]);
		DrawSquare(to_y - 1, to_x,
			   boards[currentMove][to_y - 1][to_x]);
		if (gameMode == PlayFromGameFile)
			sprintf(message, "%d. %s", move, yytext);
		else
			sprintf(message, "%d. %c%c ep", move,
				from_x + 'a', to_x + 'a');
	} else if (from_y == 7 && from_x == 4
			&& boards[currentMove][from_y][from_x] == BlackKing
			&& to_y == 7 && to_x == 6) {
		*move_type = BlackKingSideCastle;
		boards[currentMove][7][4] = EmptySquare;
		boards[currentMove][7][5] = BlackRook;
		boards[currentMove][7][6] = BlackKing;
		boards[currentMove][7][7] = EmptySquare;
		DrawSquare(7, 7, boards[currentMove][7][7]);
		DrawSquare(7, 6, boards[currentMove][7][6]);
		DrawSquare(7, 5, boards[currentMove][7][5]);
		DrawSquare(7, 4, boards[currentMove][7][4]);
		if (gameMode == PlayFromGameFile)
			sprintf(message, "%d. ... %s", move, yytext);
		else
			sprintf(message, "%d. ... 0-0", move);
	} else if (from_y == 7 && from_x == 4
			&& boards[currentMove][from_y][from_x] == BlackKing
			&& to_y == 7 && to_x == 2) {
		*move_type = BlackQueenSideCastle;
		boards[currentMove][7][0] = EmptySquare;
		boards[currentMove][7][2] = BlackKing;
		boards[currentMove][7][3] = BlackRook;
		boards[currentMove][7][4] = EmptySquare;
		DrawSquare(7, 0, boards[currentMove][7][0]);
		DrawSquare(7, 2, boards[currentMove][7][2]);
		DrawSquare(7, 3, boards[currentMove][7][3]);
		DrawSquare(7, 4, boards[currentMove][7][4]);
		if (gameMode == PlayFromGameFile)
			sprintf(message, "%d. ... %s", move, yytext);
		else
			sprintf(message, "%d. ... 0-0-0", move);
	} else if (from_y == 1
		   && boards[currentMove][from_y][from_x] == BlackPawn
		   && to_y == 0) {
		/* black pawn promotion */
		char pieceChar;

		switch (*move_type) {
		      case BlackPromotionQueen:
			boards[currentMove][0][to_x] = BlackQueen;
			pieceChar = 'Q';
			break;
		      case BlackPromotionRook:
        		boards[currentMove][0][to_x] = BlackRook;
			pieceChar = 'R';
			break;
		      case BlackPromotionBishop:
        		boards[currentMove][0][to_x] = BlackBishop;
			pieceChar = 'B';
			break;
		      case BlackPromotionKnight:
        		boards[currentMove][0][to_x] = BlackKnight;
			pieceChar = 'N';
			break;
		      default:
			*move_type = BlackPromotionQueen;
        		boards[currentMove][0][to_x] = BlackQueen;
			pieceChar = 'Q';
			break;
		}

		boards[currentMove][1][from_x] = EmptySquare;
		DrawSquare(1, from_x, boards[currentMove][1][from_x]);
		DrawSquare(0, to_x, boards[currentMove][0][to_x]);
		if (gameMode == PlayFromGameFile)
			sprintf(message, "%d. ... %s", move, yytext);
		else
			sprintf(message, "%d. ... %c2%c1(%c)", 
				move, from_x + 'a', to_x + 'a', pieceChar);

	} else if ((from_y == 3)
			&& (to_x != from_x)
			&& (boards[currentMove][from_y][from_x] == BlackPawn)
			&& (boards[currentMove][to_y][to_x] == EmptySquare)) {
		*move_type = BlackCapturesEnPassant;
		boards[currentMove][from_y][from_x] = EmptySquare;
		boards[currentMove][to_y][to_x] = BlackPawn;
		boards[currentMove][to_y + 1][to_x] = EmptySquare;
		DrawSquare(from_y, from_x,
			   boards[currentMove][from_y][from_x]);
		DrawSquare(to_y, to_x, boards[currentMove][to_y][to_x]);
		DrawSquare(to_y + 1, to_x,
			   boards[currentMove][to_y + 1][to_x]);
		if (gameMode == PlayFromGameFile)
			sprintf(message, "%d. ... %s", move, yytext);
		else
			sprintf(message, "%d. ... %c%c ep",
				move, from_x + 'a', to_x + 'a');
	} else {
		*move_type = NormalMove;
		boards[currentMove][to_y][to_x] =
		  boards[currentMove][from_y][from_x];
		boards[currentMove][from_y][from_x] = EmptySquare;
		DrawSquare(to_y, to_x, boards[currentMove][to_y][to_x]);
		DrawSquare(from_y, from_x,
			   boards[currentMove][from_y][from_x]);
		if (gameMode == PlayFromGameFile)
			sprintf(message, "%d. %s%s", move,
				WHITE_ON_MOVE ? "... " : "", yytext);
		else
			sprintf(message, "%d. %s%c%c%c%c",
				move, WHITE_ON_MOVE ? "... " : "",
				from_x + 'a', from_y + '1',
				to_x + 'a', to_y + '1');
	}

	DisplayMessage(message);
	strcpy(parseList[currentMove - 1], message);
	XSync(xDisplay, False);
}

void
InitChessProgram(host_name, program_name, pid, to, from, xid)
	char *host_name, *program_name;
	int *pid;
	FILE **to, **from;
	XtIntervalId *xid;
{
	char arg_buf[10];
	char *arg1, *arg2;
	char response[MSG_SIZ];
	int to_prog[2], from_prog[2];
	FILE *from_fp, *to_fp;
	int dummy_source;
	XtInputId dummy_id;

	if (appData.noChessProgram) return;

#if	SYSTEM_FIVE || SYSV
	if ((ttyname = PseudoTTY(&to_prog[1])) == NULL) {
		perror(program_name);
		exit(1);
	}

	from_prog[0] = to_prog[1];
	to_prog[0] = from_prog[1] = open(ttyname, O_RDWR, 0);
#else
	signal(SIGPIPE, CatchPipeSignal);
	pipe(to_prog);
	pipe(from_prog);
#endif

	if ((*pid = fork()) == 0) {
#if	!SYSTEM_FIVE && !SYSV
		signal(SIGPIPE, CatchPipeSignal);
#ifdef	O_NDELAY
		fcntl(from_prog[0], F_SETFL, O_NDELAY);
#endif
#endif
		dup2(to_prog[0], 0);
		dup2(from_prog[1], 1);
		close(to_prog[0]);
		close(to_prog[1]);
		close(from_prog[0]);
		close(from_prog[1]);
		dup2(1, fileno(stderr));	/* force stderr to the pipe */
		if (appData.searchTime != NULL) {
			arg1 = "1";
			arg2 = appData.searchTime;
		} else if (appData.searchLevel > 0) {
			sprintf(arg_buf, "%d", appData.searchLevel);
			arg1 = arg_buf;
			arg2 = (char *) NULL;
		} else {
			sprintf(arg_buf, "%d", appData.movesPerSession);
			arg1 = arg_buf;
			arg2 = appData.timeControl;
		}
		if (strcmp(host_name, "localhost") == 0) {
			execlp(program_name, program_name, arg1, arg2,
				(char *) NULL);
		} else {
			execlp(appData.remoteShell, appData.remoteShell,
				host_name, program_name, arg1, arg2,
				(char *) NULL);
		}

		perror(program_name);
		exit(1);
	}

	close(to_prog[0]);
	close(from_prog[1]);

	*from = from_fp = fdopen(from_prog[0], "r");
	*to = to_fp = fdopen(to_prog[1], "w");
	setbuf(from_fp, NULL); setbuf(to_fp, NULL);

	ReceiveFromProgram(from_fp, &dummy_source, &dummy_id); /*"Chess"*/

	*xid = XtAppAddInput(appContext, fileno(from_fp), 
			     (XtPointer)XtInputReadMask,
			     (XtInputCallbackProc)ReceiveFromProgram, 
			     (XtPointer)from_fp);

	SendToProgram(appData.initString, *to);
}

void
ShutdownChessPrograms(why)
	char *why;
{
	lastGameMode = gameMode;
	gameMode = EndOfGame;
	ModeHighlight();
	CopyBoard(boards[currentMove + 1], boards[currentMove]);
	strncpy(moveList[currentMove], why, MOVE_LEN);
	moveList[currentMove][MOVE_LEN-1] = '\000';
	currentMove++;
	DisplayMessage(why);

	if (readGameXID != NULL)
		XtRemoveTimeOut(readGameXID);
	readGameXID = NULL;

	if (firstProgramPID != 0) {
		fclose(fromFirstProgFP);
		fclose(toFirstProgFP);
		fromFirstProgFP = toFirstProgFP = NULL;
		if (kill(firstProgramPID, SIGTERM)==0) 
		  wait((union wait *)0);

	}
	firstProgramPID = 0;

	if (firstProgramXID != NULL)
		XtRemoveInput(firstProgramXID);
	firstProgramXID = NULL;

	if (secondProgramPID != 0) {
		fclose(fromSecondProgFP);
		fclose(toSecondProgFP);
		fromSecondProgFP = toSecondProgFP = NULL;
		if (kill(secondProgramPID, SIGTERM)==0) 
		  wait((union wait *)0);
	}
	secondProgramPID = 0;

	if (secondProgramXID != NULL)
		XtRemoveInput(secondProgramXID);
	secondProgramXID = NULL;

	DisplayClocks(StopTimers);

 	if(matchMode != MatchFalse) {
 		if (appData.saveGameFile[0] != '\0') 
		  SaveGameProc(appData.saveGameFile);
		exit(0);
 	}
}

void
CommentPopUp(label)
	char *label;
{
	Arg args[2];
	Position x, y;
	Dimension bw_width, pw_width;

	if (commentUp) {
		XtPopdown(commentShell);
		XtDestroyWidget(commentShell);
		commentUp = False;
	}

	DisplayMessage("Comment");

	XtSetArg(args[0], XtNwidth, &bw_width);
	XtGetValues(formWidget, args, 1);

	XtSetArg(args[0], XtNresizable, True);
	XtSetArg(args[1], XtNwidth, bw_width - 8);

	commentShell = XtCreatePopupShell("Comment",
		transientShellWidgetClass, commandsWidget, args, 2);

	XtSetArg(args[0], XtNlabel, label);

	(void) XtCreateManagedWidget("commentLabel", labelWidgetClass,
		commentShell, args, 1);

	XtRealizeWidget(commentShell);

	XtSetArg(args[0], XtNwidth, &pw_width);
	XtGetValues(commentShell, args, 1);

	XtTranslateCoords(shellWidget, (bw_width - pw_width) / 2, -50, &x, &y);

	XtSetArg(args[0], XtNx, x);
	XtSetArg(args[1], XtNy, y);
	XtSetValues(commentShell, args, 2);

	XtPopup(commentShell, XtGrabNone);
	commentUp = True;
}

void
FileNamePopUp(label, proc)
	char *label;
	void (*proc) P((char *name));
{
	Arg args[2];
	Widget popup, dialog;
	Position x, y;
	Dimension bw_width, pw_width;

	fileProc = (XtCallbackProc) proc;

	XtSetArg(args[0], XtNwidth, &bw_width);
	XtGetValues(boardWidget, args, 1);

	XtSetArg(args[0], XtNresizable, True);
	XtSetArg(args[1], XtNwidth, DIALOG_SIZE);

	popup = XtCreatePopupShell("File Name Prompt",
		transientShellWidgetClass, commandsWidget, args, 2);

	XtSetArg(args[0], XtNlabel, label);
	XtSetArg(args[1], XtNvalue, "");

	dialog = XtCreateManagedWidget("dialog", dialogWidgetClass,
		popup, args, 2);

	XawDialogAddButton(dialog, "ok", FileNameCallback, (XtPointer) dialog);
	XawDialogAddButton(dialog, "cancel", FileNameCallback,
			   (XtPointer) dialog);

	XtRealizeWidget(popup);

	XtSetArg(args[0], XtNwidth, &pw_width);
	XtGetValues(popup, args, 1);

	XtTranslateCoords(boardWidget, (bw_width - pw_width) / 2, 10, &x, &y);

	XtSetArg(args[0], XtNx, x);
	XtSetArg(args[1], XtNy, y);
	XtSetValues(popup, args, 2);

	XtPopup(popup, XtGrabExclusive);
	filenameUp = True;

	XtSetKeyboardFocus(shellWidget, popup);
}

void
FileNameCallback(w, client_data, call_data)
	Widget w;
	XtPointer client_data, call_data;
{
	String name;
	Arg args[1];

	XtSetArg(args[0], XtNlabel, &name);
	XtGetValues(w, args, 1);

	if (strcmp(name, "cancel") == 0) {
		XtPopdown(w = XtParent(XtParent(w)));
		XtDestroyWidget(w);
		filenameUp = False;
		ModeHighlight();
		return;
	}

	FileNameAction(w, NULL);
}

void
FileNameAction(w, event)
	Widget w;
	XEvent *event;
{
	char buf[MSG_SIZ];
	String name;

	name = XawDialogGetValueString(w = XtParent(w));

	if ((name != NULL) && (*name != '\0')) {
		strcpy(buf, name);
		XtPopdown(w = XtParent(w));
		XtDestroyWidget(w);
		filenameUp = False;
		(*fileProc)(buf);  /* I can't see a way not
				      to use a global here */
		ModeHighlight();
		return;
	}

	XtPopdown(w = XtParent(w));
	XtDestroyWidget(w);
	filenameUp = False;
	ModeHighlight();
}

typedef struct {
	ChessSquare piece;
	int to_x, to_y;
} PromotionMoveInfo;

static PromotionMoveInfo pmi;  /*making this global is gross */

void
PromotionPopUp(piece, to_x, to_y)
         ChessSquare piece;
         int to_x, to_y;
{
	Arg args[2];
	Widget dialog;
	Position x, y;
	Dimension bw_width, bw_height, pw_width, pw_height;

	pmi.piece = piece;
	pmi.to_x = to_x;
	pmi.to_y = to_y;

	XtSetArg(args[0], XtNwidth, &bw_width);
	XtSetArg(args[1], XtNheight, &bw_height);
	XtGetValues(boardWidget, args, 2);

	XtSetArg(args[0], XtNresizable, True);
	promotionShell = XtCreatePopupShell("Promotion",
		transientShellWidgetClass, commandsWidget, args, 1);

	XtSetArg(args[0], XtNlabel, "Promote pawn to what?");
	dialog = XtCreateManagedWidget("promotion", dialogWidgetClass,
		promotionShell, args, 1);

	XawDialogAddButton(dialog, "Queen", PromotionCallback, 
			   (XtPointer) dialog);
	XawDialogAddButton(dialog, "Rook", PromotionCallback, 
			   (XtPointer) dialog);
	XawDialogAddButton(dialog, "Bishop", PromotionCallback, 
			   (XtPointer) dialog);
	XawDialogAddButton(dialog, "Knight", PromotionCallback, 
			   (XtPointer) dialog);
	XawDialogAddButton(dialog, "cancel", PromotionCallback, 
			   (XtPointer) dialog);

	XtRealizeWidget(promotionShell);

	XtSetArg(args[0], XtNwidth, &pw_width);
	XtSetArg(args[1], XtNheight, &pw_height);
	XtGetValues(promotionShell, args, 2);

	XtTranslateCoords(boardWidget, (bw_width - pw_width) / 2,
			  LINE_GAP + squareSize/3 +
			  ((piece == WhitePawn) ^ (flipView) ?
			   0 : 6*(squareSize + LINE_GAP)),
			  &x, &y);

	XtSetArg(args[0], XtNx, x);
	XtSetArg(args[1], XtNy, y);
	XtSetValues(promotionShell, args, 2);

	XtPopup(promotionShell, XtGrabNone);

	promotionUp = True;
}

void
PromotionCallback(w, client_data, call_data)
	Widget w;
	XtPointer client_data, call_data;
{
	String name;
	Arg args[1];
	ChessMove move_type;

	XtSetArg(args[0], XtNlabel, &name);
	XtGetValues(w, args, 1);

	XtPopdown(w = XtParent(XtParent(w)));
	XtDestroyWidget(w);
	promotionUp = False;

	if (fromX == -1) return;

	if (strcmp(name, "Queen") == 0) {
		if (pmi.piece == WhitePawn)
		  move_type = WhitePromotionQueen;
		else
		  move_type = BlackPromotionQueen;
	} else if (strcmp(name, "Rook") == 0) {
		if (pmi.piece == WhitePawn)
		  move_type = WhitePromotionRook;
		else
		  move_type = BlackPromotionRook;
	} else if (strcmp(name, "Bishop") == 0) {
		if (pmi.piece == WhitePawn)
		  move_type = WhitePromotionBishop;
		else
		  move_type = BlackPromotionBishop;
	} else if (strcmp(name, "Knight") == 0) {
		if (pmi.piece == WhitePawn)
		  move_type = WhitePromotionKnight;
		else
		  move_type = BlackPromotionKnight;
	} else /* strcmp(name, "cancel") == 0 */ {
		fromX = fromY = -1;
		return;
	}

	MakeMove(&move_type, fromX, fromY, pmi.to_x, pmi.to_y);
	FinishUserMove(move_type, pmi.to_x, pmi.to_y);
}

void
SelectCommand(w, client_data, call_data)
	Widget w;
	XtPointer client_data, call_data;
{
	XawListReturnStruct *list_return = XawListShowCurrent(w);

	fromX = fromY = -1;

	if (promotionUp) {
		XtPopdown(promotionShell);
		XtDestroyWidget(promotionShell);
		promotionUp = False;
	}

	if ((gameMode == PauseGame)
		&& (list_return->list_index != (int) ButtonPause))
		PauseProc();

	switch ((Button) list_return->list_index) {
	      case ButtonQuit:
		QuitProc();
		break;
	      case ButtonBackward:
		BackwardProc();
		break;
	      case ButtonForward:
		ForwardProc();
		break;
	      case ButtonFlipView:
		FlipProc();
		break;
	      case ButtonReset:
		ResetFileProc();
		ResetProc();
		break;
	      case ButtonSaveGame:
		if (appData.saveGameFile[0] == '\0')
			FileNamePopUp("Filename for saved game?",
				      SaveGameProc);
		else
			SaveGameProc(appData.saveGameFile);
		break;
	      case ButtonSavePosition:
		if (appData.savePositionFile[0] == '\0')
			FileNamePopUp("Filename for saved position?",
				      SavePositionProc);
		else
			SavePositionProc(appData.savePositionFile);
		break;
	      case ButtonHint:
		HintProc();
		break;
#ifdef notdef
	      case ButtonSwitchSides:
		SwitchProc();
		break;
#endif /*notdef*/
	      case ButtonEditPosition:
		EditPositionProc();
		break;		
	      case ButtonLoadPosition:
		if (gameMode != BeginningOfGame) {
			DisplayMessage("Press Reset first");
			break;
		}
		if (appData.loadPositionFile == NULL)
			FileNamePopUp("Position file name?",
				VOID_PROC SetupPositionFromFileProc);
		else
			(void) SetupPositionFromFileProc(
			  appData.loadPositionFile);
		break;
	      case ButtonLoadGame:
		GameProc();
		break;
	      case ButtonMachineBlack:
		MachinePlaysBlackProc();
		break;
	      case ButtonMachineWhite:
		MachinePlaysWhiteProc();
		break;
	      case ButtonTwoMachines:
		TwoMachinesPlayProc();
		break;
	      case ButtonForceMoves:
		ForceProc();
		break;
	      case ButtonPause:
		PauseProc();
		break;
	}

	if (!filenameUp) ModeHighlight();
}

void
ModeHighlight()
{
	switch (gameMode) {
	      case BeginningOfGame:
		if (appData.noChessProgram)
		  XawListHighlight(commandsWidget, ButtonForceMoves);
		else
		  XawListHighlight(commandsWidget, ButtonMachineBlack);
		break;
	      case MachinePlaysBlack:
		XawListHighlight(commandsWidget, ButtonMachineBlack);
		break;
	      case MachinePlaysWhite:
		XawListHighlight(commandsWidget, ButtonMachineWhite);
		break;
	      case TwoMachinesPlay:
		XawListHighlight(commandsWidget, ButtonTwoMachines);
		break;
	      case ForceMoves:
	      case SetupPosition:
		XawListHighlight(commandsWidget, ButtonForceMoves);
		break;
	      case PlayFromGameFile:
		XawListHighlight(commandsWidget, ButtonLoadGame);
		break;
	      case PauseGame:
		XawListHighlight(commandsWidget, ButtonPause);
		break;
	      case EditPosition:
		XawListHighlight(commandsWidget, ButtonEditPosition);
		break;
	      case EndOfGame:
	      default:
		XawListUnhighlight(commandsWidget);
		break;
	}
}

/*
 * Button procedures
 */
void
QuitProc()
{
	ShutdownChessPrograms("Quitting");
	exit(0);
}

int
PlayFromGameFileProc(name)
	char *name;
{
	char buf[MSG_SIZ], error_buf[MSG_SIZ];
        ChessMove cm;

	if (gameMode != BeginningOfGame) {
		DisplayMessage("Press Reset first");
		return (int) False;
	}

	if (appData.loadGameFile != name) {
		if (appData.loadGameFile)
			XtFree(appData.loadGameFile);
		appData.loadGameFile = XtMalloc(strlen(name)+1);
		strcpy(appData.loadGameFile, name);
	}

	yynewfile();
	if ((gameFileFP = fopen(name, "r")) == NULL) {
		sprintf(buf, "Can't open %s", name);
		DisplayMessage(buf);
		XtFree(appData.loadGameFile);
		appData.loadGameFile = NULL;
		return (int) False;
	}

	lastGameMode = gameMode = PlayFromGameFile;
	ModeHighlight();
	InitPosition();
	DisplayClocks(StopTimers);
	if (firstProgramXID == NULL)
	  InitChessProgram(appData.firstHost, appData.firstChessProgram,
			   &firstProgramPID, &toFirstProgFP,
			   &fromFirstProgFP, &firstProgramXID);
	SendToProgram(appData.initString, toFirstProgFP);
	SendToProgram("force\n", toFirstProgFP);

	/*
	 * If the file starts with a comment line, print it.
	 */
	if (fgetc(gameFileFP) == '#') {
		fgets(buf, MSG_SIZ, gameFileFP);
		buf[strlen(buf) - 1] = '\0';
		DisplayName(buf);
	} else
		fseek(gameFileFP, 0, 0);

	/*
	 * Skip over anything that precedes an identifiable 
	 * StartGame marker, to avoid being confused by 
	 * garbage at the start of the file.  Currently 
	 * recognized StartGame markers are the move number "1"
	 * and the words "gnuchess game".
	 */
	for (;;) {
                cm = (ChessMove) yylex();
		if (cm == StartGame)
		  break;
		if (cm == (ChessMove) 0) {
			ResetFileProc();
			ResetProc();
			DisplayMessage("No game found in file");
		  	return (int) False;
		}
	}

	ReadGameFile();

	return (int) True;
}

void
ResurrectChessProgram()
     /* Restart the chess program and feed it all the moves made so far.
	Used when the user wants to back up from end of game, when gnuchess 
	has already exited.  Assumes gameMode == EndOfGame. */
{
	char buf[MSG_SIZ];
	int i;

	if (currentMove > 0)
	  currentMove--;  /* delete "White wins" or the like */

	InitChessProgram(appData.firstHost, appData.firstChessProgram,
		&firstProgramPID, &toFirstProgFP, &fromFirstProgFP,
		&firstProgramXID);
	SendToProgram(appData.initString, toFirstProgFP);
	SendToProgram("force\n", toFirstProgFP);
	gameMode = lastGameMode = ForceMoves;
	ModeHighlight();

	i = blackPlaysFirst ? 1 : 0;

	if (startedFromSetupPosition) {
		if (blackPlaysFirst)
		  SendToProgram("a3\n", toFirstProgFP);
		SendBoard(toFirstProgFP, boards[i]);
	}

	for (; i < currentMove; i++) {
		strcpy(buf, moveList[i]);
		SendToProgram(buf, toFirstProgFP);
	}

	DisplayClocks(ResetTimers);
	  /* ugh, but there's no way to tell gnuchess what the clocks
	     should read, so the best we can do is bow to its notion.
	     !!gnuchess 4.0 adds a feature that will let us fix this. */
}

void
MachinePlaysBlackProc()
{
	if (gameMode == EditPosition) EditPositionDone();

	if ((gameMode == EndOfGame) || (gameMode == PlayFromGameFile)
	    || (gameMode == TwoMachinesPlay) || appData.noChessProgram
	    || (gameMode == MachinePlaysBlack))
		return;

	if (WHITE_ON_MOVE) {
		DisplayMessage("It is not Black's turn");
		return;
	}

	lastGameMode = gameMode = MachinePlaysBlack;
	ModeHighlight();
	SendToProgram(appData.blackString, toFirstProgFP);
	DisplayClocks(DisplayTimers);
}

void
ForwardProc()
{
	char buf[MSG_SIZ];
	int i, j;

	if (gameMode == BeginningOfGame) {
		if (currentMove != 0)
			return;
		if (appData.loadGameFile != NULL) {
			if (PlayFromGameFileProc(appData.loadGameFile)) {
				XtRemoveTimeOut(readGameXID);
				readGameXID = NULL;
			}
		}
		return;
	}

	if ((gameMode == EndOfGame) || (gameMode == TwoMachinesPlay)
	    || (gameMode == EditPosition)
	    || ((currentMove == 0) && (forwardMostMove == 0)))
		return;

	if ((gameMode != PlayFromGameFile) && (currentMove >= forwardMostMove))
		return;

	if (gameMode != PauseGame) {
		PauseProc();
		if (readGameXID != NULL) {
			XtRemoveTimeOut(readGameXID);
			readGameXID = NULL;
		}
	}

	if (currentMove >= forwardMostMove) {
		ReadGameFileProc();
		return;
	}

	if ((gameMode != ForceMoves) && (gameMode != PlayFromGameFile) &&
	    (gameMode != SetupPosition)) {
		lastGameMode = gameMode = ForceMoves;
		ModeHighlight();
		SendToProgram("force\n", toFirstProgFP);
		DisplayClocks(StopTimers);
	}

	for (i = 0; i < BOARD_SIZE; i++)
	  for (j = 0; j < BOARD_SIZE; j++)
	    if (boards[currentMove + 1][i][j] != boards[currentMove][i][j])
	      DrawSquare(i, j, boards[currentMove + 1][i][j]);

	DisplayMessage(parseList[currentMove]);
	strcpy(buf, moveList[currentMove++]);
	SendToProgram(buf, toFirstProgFP);
	DisplayClocks(ReDisplayTimers);
}

void
ResetFileProc()
{
	char *buf = "";

	if (appData.loadGameFile);
		XtFree(appData.loadGameFile);
	if (appData.loadPositionFile);
		XtFree(appData.loadPositionFile);
	appData.loadGameFile = appData.loadPositionFile = NULL;
	DisplayName(buf);
	if (gameFileFP != NULL) {
		fclose(gameFileFP);
		gameFileFP = NULL;
		yynewfile();
	}
}

void
ResetProc()
{
	flipView = False;
	startedFromSetupPosition = blackPlaysFirst = False;
	matchMode = MatchFalse;
	firstMove = True;

	ShutdownChessPrograms("");
	lastGameMode = gameMode = BeginningOfGame;
	ModeHighlight();
	InitPosition();
	InitChessProgram(appData.firstHost, appData.firstChessProgram,
			 &firstProgramPID, &toFirstProgFP,
			 &fromFirstProgFP, &firstProgramXID);
	DisplayClocks(ResetTimers);
	if (commentUp) {
		XtPopdown(commentShell);
		XtDestroyWidget(commentShell);
		commentUp = False;
	}
	if (promotionUp) {
		XtPopdown(promotionShell);
		XtDestroyWidget(promotionShell);
		promotionUp = False;
	}
}

int
SetupPositionFromFileProc(name)
	char *name;
{
	char *p, line[MSG_SIZ], buf[MSG_SIZ];
	Board initial_position;
	FILE *fp;
	int i, j;

	if (gameMode != BeginningOfGame) {
		DisplayMessage("Press Reset first");
		return (int) False;
	}

	if (appData.loadPositionFile != name) {
		if (appData.loadPositionFile)
			XtFree(appData.loadPositionFile);
		appData.loadPositionFile = XtMalloc(strlen(name)+1);
		strcpy(appData.loadPositionFile, name);
	}

	if ((fp = fopen(name, "r")) == NULL) {
		sprintf(buf, "Can't open %s", name);
		DisplayMessage(buf);
		XtFree(appData.loadPositionFile);
		appData.loadPositionFile = NULL;
		return (int) False;
	}

	lastGameMode = gameMode = SetupPosition;
	ModeHighlight();
	startedFromSetupPosition = True;

	if (firstProgramXID == NULL)
	  InitChessProgram(appData.firstHost, appData.firstChessProgram,
			   &firstProgramPID, &toFirstProgFP,
			   &fromFirstProgFP, &firstProgramXID);

	/*
	 * skip header information in position file
	 */
	fgets(line, MSG_SIZ, fp);
	line[strlen(line) - 1] = '\0';
	DisplayName(line);
	fgets(line, MSG_SIZ, fp);
	fgets(line, MSG_SIZ, fp);

	for (i = BOARD_SIZE - 1; i >= 0; i--) {
		fgets(line, MSG_SIZ, fp);
		for (p = line, j = 0; j < BOARD_SIZE; p++) {
			if (*p == ' ')
				continue;
			initial_position[i][j++] = CharToPiece(*p);
		}
	}

	blackPlaysFirst = False;
        if (!feof(fp)) {
		fgets(line, MSG_SIZ, fp);
		if (strncmp(line, "black", strlen("black"))==0)
		  blackPlaysFirst = True;
	}
	fclose(fp);

	if (blackPlaysFirst) {
		CopyBoard(boards[0], initial_position);
		strcpy(moveList[0], " ...\n");
		currentMove = forwardMostMove = 1;
		CopyBoard(boards[1], initial_position);
		SendToProgram("force\na3\n", toFirstProgFP);
		SendCurrentBoard(toFirstProgFP);
		DisplayMessage("Black to play");
	} else {
		currentMove = forwardMostMove = 0;
		CopyBoard(boards[0], initial_position);
		SendCurrentBoard(toFirstProgFP);
		SendToProgram("force\n", toFirstProgFP);
		DisplayMessage("White to play");
	}

	DisplayClocks(ResetTimers);
	DrawPosition(boardWidget, (XExposeEvent *) NULL);

	return (int) True;
}

void
EditPositionProc()
{
	if (gameMode == EditPosition) return;

	ForceProc();
	if (gameMode != ForceMoves) return;

	DisplayName("<-- Press to set side to play next");
	DisplayMessage("Mouse: 1=drag, 2=white, 3=black");

	lastGameMode = gameMode = EditPosition;
	ModeHighlight();
	if (currentMove > 0)
	  CopyBoard(boards[0], boards[currentMove]);
	
	blackPlaysFirst = !WHITE_ON_MOVE;
	  /* !! later let mousing on clock change whose move it is */
	currentMove = forwardMostMove = 0;
}

void
EditPositionDone()
{
	startedFromSetupPosition = True;
	SendToProgram(appData.initString, toFirstProgFP);
	if (blackPlaysFirst) {
		strcpy(moveList[0], " ...\n");
		currentMove = forwardMostMove = 1;
		CopyBoard(boards[1], boards[0]);
		SendToProgram("force\na3\n", toFirstProgFP);
		SendCurrentBoard(toFirstProgFP);
		DisplayName(" ");
		DisplayMessage("Black to play");
	} else {
		currentMove = forwardMostMove = 0;
		SendCurrentBoard(toFirstProgFP);
		SendToProgram("force\n", toFirstProgFP);
		DisplayName(" ");
		DisplayMessage("White to play");
	}
	lastGameMode = gameMode = SetupPosition;
}

void
MachinePlaysWhiteProc()
{
	if (gameMode == EditPosition) EditPositionDone();

	if ((gameMode == EndOfGame) || (gameMode == PlayFromGameFile)
	    || (gameMode == TwoMachinesPlay) || appData.noChessProgram
	    || (gameMode == MachinePlaysWhite))
		return;

	if (!WHITE_ON_MOVE) {
		DisplayMessage("It is not White's turn");
		return;
	}

	lastGameMode = gameMode = MachinePlaysWhite;
	ModeHighlight();
	SendToProgram(appData.whiteString, toFirstProgFP);
	DisplayClocks(DisplayTimers);
}

void
BackwardProc()
{
	int i, j;

	if ((currentMove <= 0) || (gameMode == TwoMachinesPlay)
	    || (matchMode != MatchFalse) || (gameMode == EditPosition))
		return;

	if ((currentMove == 1) && blackPlaysFirst) return;

	if (gameMode == EndOfGame) ResurrectChessProgram();

	if ((WHITE_ON_MOVE && gameMode == MachinePlaysWhite)
		|| (!WHITE_ON_MOVE && gameMode == MachinePlaysBlack)) {
		DisplayMessage("Wait until your turn");
		return;
	}

	Attention(firstProgramPID);

	SendToProgram("undo\n", toFirstProgFP);

	currentMove--;
	if ((gameMode != ForceMoves) && (gameMode != PlayFromGameFile) &&
	    (gameMode != SetupPosition)) {
		SendToProgram("force\n", toFirstProgFP);
		lastGameMode = gameMode = ForceMoves;
		ModeHighlight();
		DisplayClocks(StopTimers);
        }
	DisplayClocks(ReDisplayTimers);
	DisplayMessage(parseList[currentMove - 1]);

	if (gameMode != PauseGame) {
		PauseProc();
		if (readGameXID != NULL) {
			XtRemoveTimeOut(readGameXID);
			readGameXID = NULL;
		}
	}
	for (i = 0; i < BOARD_SIZE; i++)
		for (j = 0; j < BOARD_SIZE; j++)
			if (boards[currentMove][i][j] != boards[currentMove + 1][i][j])
				DrawSquare(i, j, boards[currentMove][i][j]);
}

void
FlipProc()
{
	flipView = !flipView;
	DrawPosition(boardWidget, (XExposeEvent *) NULL);
}

void
SaveGameProc(name)
	char *name;
{
	char buf[MSG_SIZ], white_move[MSG_SIZ], black_move[MSG_SIZ];
	int i, len, move = 0;
	time_t tm;

	if ((gameFileFP = fopen(name, "w")) == NULL) {
		sprintf(buf, "Can't open %s", name);
		DisplayMessage(buf);
		return;
	}

	tm = time((time_t *) NULL);
	gethostname(buf, MSG_SIZ);

	fprintf(gameFileFP, "# %s game file -- %s", programName, ctime(&tm));
	switch (lastGameMode) {
	      case MachinePlaysWhite:
		fprintf(gameFileFP, "\t%s@%s vs. %s@%s\n",
			appData.firstChessProgram, appData.firstHost,
			getpwuid(getuid())->pw_name, buf);
		break;
	      case MachinePlaysBlack:
		fprintf(gameFileFP, "\t%s@%s vs. %s@%s\n",
			getpwuid(getuid())->pw_name, buf,
			appData.firstChessProgram, appData.firstHost);
		break;
	      case TwoMachinesPlay:
		fprintf(gameFileFP, "\t%s@%s vs. %s@%s\n",
			appData.secondChessProgram, appData.secondHost,
			appData.firstChessProgram, appData.firstHost);
		break;
	      default:
		fprintf(gameFileFP, "\n");
		break;
	}
	fprintf(gameFileFP, "\talgebraic\n");

	for (i = 0; i < currentMove;) {
		/*
		 * get rid of '\n' added to send the move to GNU Chess
		 */
		len = strlen(moveList[i]);
		strcpy(white_move, moveList[i++]);
		if (white_move[len - 1] == '\n')
			white_move[len - 1] = NULL;
		fprintf(gameFileFP, "%d. %s ", ++move, white_move);

		if (i >= currentMove) {
			fprintf(gameFileFP, "\n");
			break;
		}

		if ((len = strlen(moveList[i])) == 0) {
			fprintf(gameFileFP, "\n");
			break;
		}
		strcpy(black_move, moveList[i++]);
		if (black_move[len - 1] == '\n')
			black_move[len - 1] = NULL;
		fprintf(gameFileFP, "%s\n", black_move);
	}

	fclose(gameFileFP);
	gameFileFP = NULL;
	yynewfile();
}

#ifdef notdef  
/* unused */
void
SwitchProc()
{
	if (appData.noChessProgram) return;
	switch (gameMode) {
	default:
		return;
	case MachinePlaysWhite:
		if (WHITE_ON_MOVE) {
			DisplayMessage("Wait until your turn");
			return;
		}
		lastGameMode = gameMode = MachinePlaysBlack;
		ModeHighlight();
		break;
	case BeginningOfGame:
	case MachinePlaysBlack:
		if (!WHITE_ON_MOVE) {
			DisplayMessage("Wait until your turn");
			return;
		}
		if (currentMove == 0) {
			MachinePlaysWhiteProc();
			return;
		}
		lastGameMode = gameMode = MachinePlaysWhite;
		ModeHighlight();
		break;
	}

	Attention(firstProgramPID);
	SendToProgram("switch\n", toFirstProgFP);
}
#endif /*notdef*/

void
ForceProc()
{
	switch (gameMode) {
	      case MachinePlaysWhite:
		if (WHITE_ON_MOVE) {
			DisplayMessage("Wait until your turn");
			return;
		}
		Attention(firstProgramPID);
		SendToProgram("force\n", toFirstProgFP);
		break;
	      case MachinePlaysBlack:
		if (!WHITE_ON_MOVE) {
			DisplayMessage("Wait until your turn");
			return;
		}
		Attention(firstProgramPID);
		SendToProgram("force\n", toFirstProgFP);
		break;
	      case BeginningOfGame:
		SendToProgram("force\n", toFirstProgFP);
		break;
	      case PlayFromGameFile:
		if (readGameXID != NULL) {
			XtRemoveTimeOut(readGameXID);
			readGameXID = NULL;
		}
		if (gameFileFP != NULL) {
			fclose(gameFileFP);
			gameFileFP = NULL;
			yynewfile();
		}
		break;
	      case EndOfGame:
		ResurrectChessProgram();
		break;
	      case EditPosition:
		EditPositionDone();
		break;
	      case SetupPosition:
		break;
	      default:
		return;
	}

	lastGameMode = gameMode = ForceMoves;
	ModeHighlight();
	DisplayClocks(StopTimers);
}

void
HintProc()
{
	if (appData.noChessProgram) return;
	switch (gameMode) {
	case MachinePlaysWhite:
		if (WHITE_ON_MOVE) {
			DisplayMessage("Wait until your turn");
			return;
		}
		break;
	case BeginningOfGame:
	case MachinePlaysBlack:
		if (!WHITE_ON_MOVE) {
			DisplayMessage("Wait until your turn");
			return;
		}
		break;
	default:
		DisplayMessage("No hint available");
		return;
	}
	Attention(firstProgramPID);
	SendToProgram("hint\n", toFirstProgFP);
}

void
SavePositionProc(name)
	char *name;
{
	char buf[MSG_SIZ], host_name[MSG_SIZ];
	FILE *fp;
	time_t tm;
	int i, j;

	if ((fp = fopen(name, "w")) == NULL) {
		sprintf(buf, "Can't open %s", name);
		DisplayMessage(buf);
		return;
	}

	tm = time((time_t *) NULL);
	gethostname(host_name, MSG_SIZ);

	fprintf(fp, "%s position file -- %s", programName, ctime(&tm));
	switch (lastGameMode) {
	case MachinePlaysWhite:
		fprintf(fp, "\t%s@%s vs. %s@%s\n", appData.firstChessProgram,
			appData.firstHost, getpwuid(getuid())->pw_name,
			host_name);
		break;
	case MachinePlaysBlack:
		fprintf(fp, "\t%s@%s vs. %s@%s\n", getpwuid(getuid())->pw_name,
			host_name, appData.firstChessProgram,
			appData.firstHost);
		break;
	case TwoMachinesPlay:
		fprintf(fp, "\t%s@%s vs. %s@%s\n", appData.secondChessProgram,
			appData.secondHost, appData.firstChessProgram,
			appData.firstHost);
		break;
	default:
		fprintf(fp, "\n");
		break;
	}
	fprintf(fp, "\n");

	for (i = BOARD_SIZE - 1; i >= 0; i--) {
		for (j = 0; j < BOARD_SIZE; j++) {
			fprintf(fp, "%c",
				pieceToChar[(int) boards[currentMove][i][j]]);
			fputc(j == BOARD_SIZE - 1 ? '\n' : ' ', fp);
		}
	}

	if ((gameMode == EditPosition) ? !blackPlaysFirst : WHITE_ON_MOVE)
	  fprintf(fp, "white to play\n");
	else
	  fprintf(fp, "black to play\n");

	fclose(fp);
}

void
TwoMachinesPlayProc()
{
	int i;
	MatchMode matchKind;

	if ((gameMode == EndOfGame) || (gameMode == TwoMachinesPlay)
	    || appData.noChessProgram)
		return;

	if (matchMode == MatchFalse) {
		switch (gameMode) {
		      case PauseGame:
		      case PlayFromGameFile:
			return;
		      case MachinePlaysWhite:
		      case MachinePlaysBlack:
			ForceProc();
			if (gameMode != ForceMoves) return;
			matchKind = MatchOpening;
			break;
		      case ForceMoves:
			matchKind = MatchOpening;
			break;
		      case SetupPosition:
			matchKind = MatchPosition;
			break;
		      case EditPosition:
			EditPositionDone();
			matchKind = MatchPosition;
			break;
		      case BeginningOfGame:
		      default:
			matchKind = MatchInit;
			break;
		}
	} else {
		matchKind = matchMode;
	}

	flipView = False;
	firstMove = False;
	DisplayClocks(ResetTimers);
	DisplayClocks(ReStartTimers);

	switch (matchKind) {
	      case MatchOpening:
		if (firstProgramXID == NULL) {
			if (appData.loadGameFile == NULL) {
				DisplayMessage("Select game file first");
				return;
			}
			InitChessProgram(appData.firstHost,
					 appData.firstChessProgram,
					 &firstProgramPID, &toFirstProgFP,
					 &fromFirstProgFP, &firstProgramXID);
			if (!PlayFromGameFileProc(appData.loadGameFile)) {
				ShutdownChessPrograms("Bad game file");
				return;
			}
			DrawPosition(boardWidget, (XExposeEvent *) NULL);
		}
		InitChessProgram(appData.secondHost,
				 appData.secondChessProgram,
				 &secondProgramPID, &toSecondProgFP,
				 &fromSecondProgFP, &secondProgramXID);
		if (startedFromSetupPosition) {
			if (blackPlaysFirst) {
				i = 1;
				SendToProgram("force\na3\n", toSecondProgFP);
				SendBoard(toSecondProgFP, boards[i]);
			} else {
				i = 0;
				SendBoard(toSecondProgFP, boards[i]);
				SendToProgram("force\n", toSecondProgFP);
			}
		} else {
			i = 0;
			SendToProgram("force\n", toSecondProgFP);
		}
		for (; i < currentMove; i++)
			SendToProgram(moveList[i], toSecondProgFP);
		lastGameMode = gameMode = TwoMachinesPlay;
		ModeHighlight();
		firstMove = True;
		if (WHITE_ON_MOVE)
		  SendToProgram(appData.whiteString, toSecondProgFP);
		else
		  SendToProgram(appData.blackString, toFirstProgFP);
		break;

	      case MatchPosition:
		if (firstProgramXID == NULL) {
			if (appData.loadPositionFile == NULL) {
				DisplayMessage("Select position file first");
				return;
			}
			InitChessProgram(appData.firstHost,
					 appData.firstChessProgram,
					 &firstProgramPID, &toFirstProgFP,
					 &fromFirstProgFP, &firstProgramXID);
			if (!SetupPositionFromFileProc(
  			          appData.loadPositionFile))
				return;
		}
		InitChessProgram(appData.secondHost,
				 appData.secondChessProgram,
				 &secondProgramPID, &toSecondProgFP,
				 &fromSecondProgFP, &secondProgramXID);
		if (blackPlaysFirst)
		  SendToProgram("force\na3\n", toSecondProgFP);
		SendCurrentBoard(toSecondProgFP);
		lastGameMode = gameMode = TwoMachinesPlay;
		ModeHighlight();
		firstMove = True;
		if (WHITE_ON_MOVE)
		  SendToProgram(appData.whiteString, toSecondProgFP);
		else
		  SendToProgram(appData.blackString, toFirstProgFP);
		break;

	      case MatchInit:
		InitPosition();
		if (firstProgramXID == NULL)
			InitChessProgram(appData.firstHost,
					 appData.firstChessProgram,
					 &firstProgramPID, &toFirstProgFP,
					 &fromFirstProgFP, &firstProgramXID);
		InitChessProgram(appData.secondHost,
				 appData.secondChessProgram,
				 &secondProgramPID, &toSecondProgFP,
				 &fromSecondProgFP, &secondProgramXID);
		lastGameMode = gameMode = TwoMachinesPlay;
		ModeHighlight();
		SendToProgram(appData.whiteString, toSecondProgFP);

	      default:
		break;
	}
}

void
PauseProc()
{
	static GameMode previous_mode = PauseGame;

	switch (gameMode) {
	      case ForceMoves:
	      case EndOfGame:
	      case SetupPosition:
	      case EditPosition:
	      default:
		return;
	      case PauseGame:
		gameMode = previous_mode;
		ModeHighlight();
		previous_mode = PauseGame;
		DisplayClocks(ReStartTimers);
		DisplayMessage("");
		break;
	      case PlayFromGameFile:
		if (readGameXID == NULL) {
			readGameXID =
			  XtAppAddTimeOut(appContext,
					  (int) (1000 * appData.timeDelay),
					  ReadGameFile, NULL);
		} else {
			XtRemoveTimeOut(readGameXID);
			readGameXID = NULL;
		}
		DisplayMessage("Pausing");
		break;
	      case BeginningOfGame:
	      case MachinePlaysWhite:
	      case MachinePlaysBlack:
	      case TwoMachinesPlay:
		if (currentMove == 0)	/* don't pause if no one has moved */
			return;
		previous_mode = gameMode;
		gameMode = PauseGame;
		ModeHighlight();
		DisplayClocks(StopTimers);
		DisplayMessage("Pausing");
		break;
	}
}

void
GameProc()
{
	if (gameMode != BeginningOfGame) {
		DisplayMessage("Press Reset first");
		return;
	}
	if (appData.loadGameFile == NULL)
		FileNamePopUp("Game file name?",
			      VOID_PROC PlayFromGameFileProc);
	else
		(void) PlayFromGameFileProc(appData.loadGameFile);
}

void
Iconify()
{
	Arg args[1];


	fromX = fromY = -1;

	XtSetArg(args[0], XtNiconic, True);
	XtSetValues(shellWidget, args, 1);
}

void
SendToProgram(message, fp)
	char *message;
	FILE *fp;
{
	if (fp == NULL) return;
	lastMsgFP = fp;

	if (xboardDebug)
		fprintf(stderr, "Sending to %s: %s\n",
			fp == toFirstProgFP ? "first" : "second", message);

	if (message[strlen(message) - 1] != '\n')
		fprintf(fp, "\n%s\n", message);
	else
		fputs(message, fp);
	fflush(fp);
}

void
ReceiveFromProgram(fp, source, id)
	FILE *fp;
        int *source;
        XtInputId *id;
{
	char message[MSG_SIZ], *end_str;

	if (fgets(message, MSG_SIZ, fp) == NULL) {
		sprintf(message,
			"%s chess program (%s) exited unexpectedly",
			fp == fromFirstProgFP ? "first" : "second",
			fp == fromFirstProgFP ? appData.firstChessProgram
			    : appData.secondChessProgram);
		fprintf(stderr, "%s: %s\n", programName, message);
		ShutdownChessPrograms(message);
		return;
	}

	if ((end_str = (char *) strchr(message, '\r')) != NULL)
		*end_str = '\0';
	if ((end_str = (char *) strchr(message, '\n')) != NULL)
		*end_str = '\0';

	if (xboardDebug)
		fprintf(stderr, "Received from %s: %s\n",
			fp == fromFirstProgFP ? "first" : "second", message);
	HandleMachineMove(message, fp);
}

void
DisplayMessage(message)
	char *message;
{
	Arg arg;

	XtSetArg(arg, XtNlabel, message);
	XtSetValues(messageWidget, &arg, 1);
}

void
DisplayName(name)
	char *name;
{
	Arg arg;

	XtSetArg(arg, XtNlabel, name);
	XtSetValues(nameWidget, &arg, 1);
}

/*
 * This routine sends a SIGINT (^C interrupt) to gnuchess to awaken it
 * if it might be busy thinking on our time.  This isn't needed in
 * any released versions of gnuchess that I know of, but was needed
 * in beta versions of 4.0, patchlevel 19 and below.  I've kept the
 * code here in case someone needs it later.  Similar code could be
 * used to stop gnuchess and make it move immediately when it is 
 * thinking about its own move; this could be useful if we want to
 * make Backward or ForceMoves work while gnuchess is thinking. --t.mann
 */
void
Attention(pid)
     int pid;
{
#ifdef ATTENTION
	if (appData.noChessProgram || (pid == 0)) return;
	switch (gameMode) {
	      case MachinePlaysWhite:
	      case MachinePlaysBlack:
	      case TwoMachinesPlay:
		if (currentMove > (blackPlaysFirst ? 2 : 1)) {
			if (xboardDebug)
			  fprintf(stderr, "Sending SIGINT to %s\n",
			    pid == firstProgramPID ? "first" : "second");

			(void) kill(pid, SIGINT); /* stop it thinking */
		}
		break;
	}
#endif /*ATTENTION*/
}

/*
 * DisplayClocks manages the game clocks.
 *
 * In tournament play, black starts the clock and then white makes a move.
 * We give the human user a slight advantage if he is playing white---the
 * clocks don't run until he makes his first move, so it takes zero time.
 * Also, DisplayClocks doesn't account for network lag so it could get
 * out of sync with GNU Chess's clock -- but then, referees are always right.
 */
void
DisplayClocks(clock_mode)
	int clock_mode;
{
	switch (clock_mode) {
	case ResetTimers:
		if (timerXID != NULL) {
			XtRemoveTimeOut(timerXID);
			timerXID = NULL;
		}

		whiteTimeRemaining = timeControl;
		DisplayTimerLabel(whiteTimerWidget, "White",
				  whiteTimeRemaining);

		blackTimeRemaining = timeControl;
		DisplayTimerLabel(blackTimerWidget, "Black",
				  blackTimeRemaining);
		break;
        case DisplayTimers:
		if (gameMode == PauseGame)
			return;
		if (appData.clockMode) {
			if (WHITE_ON_MOVE) {
				whiteTimeRemaining--;
				if (whiteTimeRemaining <= 0) {
					whiteTimeRemaining = 0;
					DisplayMessage("White's flag dropped");
				}
			} else {
				blackTimeRemaining--;
				if (blackTimeRemaining <= 0) {
					blackTimeRemaining = 0;
					DisplayMessage("Black's flag dropped");
				}
			}
		}

		DisplayTimerLabel(whiteTimerWidget, "White",
				  whiteTimeRemaining);
		DisplayTimerLabel(blackTimerWidget, "Black",
				  blackTimeRemaining);

		/*
		 * reset clocks when time control is acheived
		 */
		if (appData.clockMode && (currentMove != 0)
		    && ((currentMove % (appData.movesPerSession * 2)) == 0)
		    && (whiteTimeRemaining > 0) && (blackTimeRemaining > 0)) {

			whiteTimeRemaining = timeControl;
			blackTimeRemaining = timeControl;
		}

		if (timerXID != NULL) {
			XtRemoveTimeOut(timerXID);
			timerXID = NULL;
		}
		if (appData.clockMode)
			timerXID = XtAppAddTimeOut(appContext,
				1000, DisplayClocks, (XtPointer)DisplayTimers);
		break;
	case ReDisplayTimers:
		DisplayTimerLabel(whiteTimerWidget, "White",
				  whiteTimeRemaining);
		DisplayTimerLabel(blackTimerWidget, "Black",
				  blackTimeRemaining);
		break;
	case StopTimers:
		if (timerXID == NULL)
			return;
		XtRemoveTimeOut(timerXID);
		timerXID = NULL;
		break;
	case ReStartTimers:
		if (timerXID != NULL)
			return;

		DisplayTimerLabel(whiteTimerWidget, "White",
				  whiteTimeRemaining);
		DisplayTimerLabel(blackTimerWidget, "Black",
				  blackTimeRemaining);

		if (appData.clockMode)
			timerXID = XtAppAddTimeOut(appContext,
				1000, DisplayClocks, (XtPointer)DisplayTimers);
		break;
	}
}

void
DisplayTimerLabel(w, color, timer)
	Widget w;
	char *color;
	time_t timer;
{
	char buf[MSG_SIZ];
	Arg args[3];

	if (appData.clockMode) {
		sprintf(buf, "%s: %s", color, TimeString(timer));
		XtSetArg(args[0], XtNlabel, buf);
	} else
		XtSetArg(args[0], XtNlabel, color);

	if (((color[0] == 'B') && WHITE_ON_MOVE)
		|| ((color[0] == 'W') && !WHITE_ON_MOVE)) {
		XtSetArg(args[1], XtNbackground, timerForegroundPixel);
		XtSetArg(args[2], XtNforeground, timerBackgroundPixel);
	} else {
		XtSetArg(args[1], XtNbackground, timerBackgroundPixel);
		XtSetArg(args[2], XtNforeground, timerForegroundPixel);
	}

	XtSetValues(w, args, 3);
}

char *
TimeString(tm)
	time_t tm;
{
	int second, minute, hour, day;
	static char buf[32];

	if (tm >= (60 * 60 * 24)) {
		day = (int) (tm / (60 * 60 * 24));
		tm -= day * 60 * 60 * 24;
	} else
		day = 0;

	if (tm >= (60 * 60)) {
		hour = (int) (tm / (60 * 60));
		tm -= hour * 60 * 60;
	} else
		hour = 0;

	if (tm >= 60) {
		minute = (int) (tm / 60);
		tm -= minute * 60;
	} else
		minute = 0;

	second = tm % 60;

	if (day > 0)
		sprintf(buf, " %d:%02d:%02d:%02d ", day, hour, minute, second);
	else if (hour > 0)
		sprintf(buf, " %d:%02d:%02d ", hour, minute, second);
	else
		sprintf(buf, " %2d:%02d ", minute, second);

	return buf;
}

void
Usage()
{
	fprintf(stderr, "Usage: %s\n", programName);
	fprintf(stderr, "\tstandard Xt options\n");
	fprintf(stderr, "\t-iconic\n");
	fprintf(stderr, "\t-tc or -timeControl minutes[:seconds]\n");
	fprintf(stderr, "\t-mps or -movesPerSession moves\n");
	fprintf(stderr, "\t-st or -searchTime minutes[:seconds]\n");
	fprintf(stderr, "\t-sl or -searchLevel number\n");
	fprintf(stderr, "\t-clock or -clockMode (True | False)\n");
	fprintf(stderr, "\t-td or -timeDelay seconds\n");
	fprintf(stderr, "\t-ncp or -noChessProgram (True | False)\n");
	fprintf(stderr, "\t-fcp or -firstChessProgram program_name\n");
	fprintf(stderr, "\t-scp or -secondChessProgram program_name\n");
	fprintf(stderr, "\t-fh or -firstHost host_name\n");
	fprintf(stderr, "\t-sh or -secondHost host_name\n");
	fprintf(stderr, "\t-rsh or -remoteShell shell_name\n");
	fprintf(stderr,
		"\t-mm or -matchMode (False | Init | Position | Opening)\n");
	fprintf(stderr, "\t-lgf or -loadGameFile file_name\n");
	fprintf(stderr, "\t-lpf or -loadPositionFile file_name\n");
	fprintf(stderr, "\t-sgf or -saveGameFile file_name\n");
	fprintf(stderr, "\t-spf or -savePositionFile file_name\n");
	fprintf(stderr, "\t-size or -boardSize (Large | Medium | Small)\n");
	fprintf(stderr, "\t-mono or -monoMode (True | False)\n");
	fprintf(stderr, "\t-wpc or -whitePieceColor color\n");
	fprintf(stderr, "\t-bpc or -blackPieceColor color\n");
	fprintf(stderr, "\t-lsc or -lightSquareColor color\n");
	fprintf(stderr, "\t-dsc or -darkSquareColor color\n");
	fprintf(stderr, "\t-spb or -solidPawnBitmap file_name\n");
	fprintf(stderr, "\t-srb or -solidRookBitmap file_name\n");
	fprintf(stderr, "\t-sbb or -solidBishopBitmap file_name\n");
	fprintf(stderr, "\t-snb or -solidKnightBitmap file_name\n");
	fprintf(stderr, "\t-sqb or -solidQueenBitmap file_name\n");
	fprintf(stderr, "\t-skb or -solidKingBitmap file_name\n");
	fprintf(stderr, "\t-opb or -outlinePawnBitmap file_name\n");
	fprintf(stderr, "\t-orb or -outlineRookBitmap file_name\n");
	fprintf(stderr, "\t-obb or -outlineBishopBitmap file_name\n");
	fprintf(stderr, "\t-onb or -outlineKnightBitmap file_name\n");
	fprintf(stderr, "\t-oqb or -outlineQueenBitmap file_name\n");
	fprintf(stderr, "\t-okb or -outlineKingBitmap file_name\n");
	fprintf(stderr, "\t-debug or -debugMode (True | False)\n");
	exit(2);
}

/*
 * This is necessary because some C libraries aren't ANSI C compliant yet.
 */
char *
StrStr(string, match)
	char *string, *match;
{
	int i, length;

	length = strlen(match);

	for (i = strlen(string) - length; i >= 0; i--, string++)
		if (!strncmp(match, string, (size_t) length))
			return string;

	return NULL;
}

int
StrCaseCmp(s1, s2)
     char *s1, *s2;
{
	char c1, c2;

	for (;;) {
		c1 = tolower(*s1++);
		c2 = tolower(*s2++);
		if (c1 > c2) return 1;
		if (c1 < c2) return -1;
		if (c1 == '\000') return 0;
	}
}


#if	SYSTEM_FIVE || SYSV
char *
PseudoTTY(ptyv)
	int *ptyv;
{
	struct stat stb;
	int c, i;

	for (c = 'p'; c <= 'z'; c++)
		for (i = 0; i < 16; i++) {
#ifdef	HPUX
			sprintf(ptyname, "/dev/ptym/pty%c%x", c, i);
#else
#ifdef	RTU
			sprintf(ptyname, "/dev/pty%x", i);
#else
			sprintf(ptyname, "/dev/pty%c%x", c, i);
#endif
#endif

#ifdef	IRIS
			*ptyv = open("/dev/ptc", O_RDWR | O_NDELAY, 0);
			if (*ptyv < 0)
				return NULL;
			if (fstat(*ptyv, &stb) < 0)
				return NULL;
#else
			if (stat(ptyname, &stb) < 0)
				return NULL;
			*ptyv = open(ptyname, O_RDWR | O_NDELAY, 0);
#endif

			if (*ptyv >= 0) {
#ifdef	HPUX
				sprintf(ptyname, "/dev/pty/tty%c%x", c, i);
#else
#ifdef	RTU
				sprintf(ptyname, "/dev/ttyp%x", i);
#else
#ifdef	IRIS
				sprintf(ptyname, "/dev/ttyq%d",
					minor(stb.st_rdev));
#else
				sprintf(ptyname, "/dev/tty%c%x", c, i);
#endif
#endif
#endif

#ifndef	UNIPLUS
				if (access(ptyname, 6) != 0) {
					close(*ptyv);
#ifdef	IRIS
					return NULL;
#else
					continue;
#endif
				}
#endif

#ifdef	FIONBIO
			{
				int on = 1;
				ioctl(*ptyv, FIONBIO, &on);
			}
#endif

#ifdef	IBMRTAIX
			signal(SIGHUP, SIG_IGN);
#endif
			return ptyname;
		}
	}

	return NULL;
}

#else

void
CatchPipeSignal()
{
	char message[MSG_SIZ];

	sprintf(message,
		"%s chess program (%s) exited unexpectedly",
		lastMsgFP == fromFirstProgFP ? "first" : "second",
		lastMsgFP == fromFirstProgFP ? appData.firstChessProgram
		    : appData.secondChessProgram);
	fprintf(stderr, "%s: %s\n", programName, message);
	ShutdownChessPrograms(message);
	return;
}
#endif
